"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import os
import platform
import pandas as pd

from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.Utils import random_user_agent

from pkscreener.classes import VERSION, Changelog, Utility
from pkscreener.classes.ArtTexts import getArtText
from pkscreener.classes.Utility import marketStatus,STD_ENCODING,lastScreened
import pkscreener.classes.Fetcher as Fetcher

# Class for managing misc image utility methods
class PKConsoleTools:
    
    fetcher = Fetcher.screenerStockDataFetcher()
    def clearScreen(userArgs=None,clearAlways=False,forceTop=False):
        timeit = 'timeit' in os.environ.keys()
        if timeit:
            return
        if "RUNNER" in os.environ.keys() or (userArgs is not None and userArgs.prodbuild):
            if userArgs is not None and userArgs.v:
                os.environ["RUNNER"]="LOCAL_RUN_SCANNER"
            return
        elif (userArgs is not None and userArgs.runintradayanalysis):
            return
        if clearAlways or OutputControls().enableMultipleLineOutput:
            if platform.system() == "Windows":
                try:
                    os.system('color 0f') # sets the background to black with white forerground
                except: # pragma: no cover
                    pass
                if clearAlways:
                    os.system("cls")
            elif "Darwin" in platform.system():
                if clearAlways:
                    os.system("clear")
                # default_logger().debug("Darwin does not work for setting background")
            else:
                try:
                    os.system('setterm -background black -foreground white -store')
                except: # pragma: no cover
                    pass
                if clearAlways:
                    os.system("clear")
            OutputControls().moveCursorToStartPosition()
        try:
            forceTop = OutputControls().enableMultipleLineOutput
            if forceTop and OutputControls().lines == 0:
                OutputControls().lines = 9
                OutputControls().moveCursorToStartPosition()
                
            if clearAlways or OutputControls().enableMultipleLineOutput:
                art = colorText.GREEN + f"{getArtText()}\n" + colorText.END + f"{marketStatus()}"
                OutputControls().printOutput(art.encode('utf-8').decode(STD_ENCODING), enableMultipleLineOutput=True)
        except KeyboardInterrupt: # pragma: no cover
            raise KeyboardInterrupt
        except Exception as e: # pragma: no cover
            default_logger().debug(e, exc_info=True)
            pass

    # Print about developers and repository
    def showDevInfo(defaultAnswer=None):
        OutputControls().printOutput("\n" + Changelog.changelog())
        devInfo = "\n[👨🏻‍💻] Developer: PK (PKScreener) 🇮🇳"
        versionInfo = "[🚦] Version: %s" % VERSION
        homePage = "[🏡] Home Page: https://github.com/pkjmesra/PKScreener\n[🤖] Telegram Bot:@nse_pkscreener_bot\n[🚨] Channel:https://t.me/PKScreener\n[💬] Discussions:https://t.me/PKScreeners"
        issuesInfo = (
            "[🚩] Read/Post Issues here: https://github.com/pkjmesra/PKScreener/issues"
        )
        communityInfo = "[📢] Join Community Discussions: https://github.com/pkjmesra/PKScreener/discussions"
        latestInfo = "[⏰] Download latest software from https://github.com/pkjmesra/PKScreener/releases/latest"
        freeInfo = "[💰] PKScreener had been free for a long time"
        donationInfo = ", but owing to cost/budgeting issues, only a basic set of features will always remain free for everyone. Consider donating to help cover the basic server costs or subscribe to premium.\n[💸] Please donate whatever you can: PKScreener@APL using UPI(India) or https://github.com/sponsors/pkjmesra 🙏🏻"
        totalDownloads = "200k+"
        respPepyTech = PKConsoleTools.fetcher.fetchURL(url="https://static.pepy.tech/badge/pkscreener",headers={'user-agent': f'{random_user_agent()}'},timeout=2)
        if respPepyTech is not None and respPepyTech.status_code == 200:
            totalDownloads = respPepyTech.text.split("</text>")[-2].split(">")[-1]
        downloadsInfo = f"[🔥] Total Downloads: {totalDownloads}"
        OutputControls().printOutput(colorText.WARN + devInfo + colorText.END)
        OutputControls().printOutput(colorText.WARN + versionInfo + colorText.END)
        OutputControls().printOutput(colorText.GREEN + downloadsInfo + colorText.END)
        OutputControls().printOutput(homePage + colorText.END)
        OutputControls().printOutput(colorText.FAIL + issuesInfo + colorText.END)
        OutputControls().printOutput(colorText.GREEN + communityInfo + colorText.END)
        OutputControls().printOutput(colorText.BLUE + latestInfo + colorText.END)
        OutputControls().printOutput(colorText.GREEN + freeInfo + colorText.END + colorText.FAIL + donationInfo + colorText.END)
        if defaultAnswer is None:
            OutputControls().takeUserInput(
                colorText.FAIL
                + "  [+] Press <Enter> to continue!"
                + colorText.END
            )
        return f"\n{Changelog.changelog()}\n\n{devInfo}\n{versionInfo}\n\n{downloadsInfo}\n{homePage}\n{issuesInfo}\n{communityInfo}\n{latestInfo}\n{freeInfo}{donationInfo}"

    # Save last screened result to pickle file
    def setLastScreenedResults(df, df_save=None, choices=None):
        try:
            finalStocks = ""
            outputFolder = os.path.join(os.getcwd(),'actions-data-scan')
            if not os.path.isdir(outputFolder):
                os.makedirs(os.path.dirname(os.path.join(os.getcwd(),f"actions-data-scan{os.sep}")), exist_ok=True)
            fileName = os.path.join(outputFolder,f"{choices}.txt")
            items = []
            needsWriting = False
            if os.path.isfile(fileName):
                if df is not None and len(df) > 0:
                    #File already exists. Let's combine because there are new stocks found
                    with open(fileName, 'r') as fe:
                        stocks = fe.read()
                        items = stocks.replace("\n","").replace("\"","").split(",")
                        stockList = sorted(list(filter(None,list(set(items)))))
                        finalStocks = ",".join(stockList)
            else:
                needsWriting = True
            if df is not None and len(df) > 0:
                with pd.option_context('mode.chained_assignment', None):
                    df.sort_values(by=["Stock"], ascending=True, inplace=True)
                    df.to_pickle(lastScreened)
                    if choices is not None and df_save is not None:
                        df_s = df_save.copy()
                        df_s.reset_index(inplace=True)
                        newStocks = df_s["Stock"].to_json(orient='records', lines=True).replace("\n","").replace("\"","").split(",")
                        items.extend(newStocks)
                        stockList = sorted(list(filter(None,list(set(items)))))
                        finalStocks = ",".join(stockList)
                        needsWriting = True
            if needsWriting:
                with open(fileName, 'w') as f:
                    f.write(finalStocks)
        except IOError as e: # pragma: no cover
            default_logger().debug(e, exc_info=True)
            OutputControls().printOutput(
                colorText.FAIL
                + f"{e}\n  [+] Failed to save recently screened result table on disk! Skipping.."
                + colorText.END
            )
        except KeyboardInterrupt: # pragma: no cover
            raise KeyboardInterrupt
        except Exception as e: # pragma: no cover
            default_logger().debug(e, exc_info=True)
            pass

    # Load last screened result to pickle file
    def getLastScreenedResults(defaultAnswer=None):
        try:
            df = pd.read_pickle(lastScreened)
            if df is not None and len(df) > 0:
                OutputControls().printOutput(
                    colorText.GREEN
                    + "\n  [+] Showing recently screened results..\n"
                    + colorText.END
                )
                df.sort_values(by=["volume" if "volume" in df.keys() else df.keys()[0]], ascending=False, inplace=True)
                OutputControls().printOutput(
                    colorText.miniTabulator().tabulate(
                        df, headers="keys", tablefmt=colorText.No_Pad_GridFormat,
                        maxcolwidths=Utility.tools.getMaxColumnWidths(df)
                    ).encode("utf-8").decode(STD_ENCODING)
                )
                OutputControls().printOutput(
                    colorText.WARN
                    + "  [+] Note: Trend calculation is based on number of recent days to screen as per your configuration."
                    + colorText.END
                )
            else:
                OutputControls().printOutput("Nothing to show here!")
        except FileNotFoundError as e: # pragma: no cover
            default_logger().debug(e, exc_info=True)
            OutputControls().printOutput(
                colorText.FAIL
                + "  [+] Failed to load recently screened result table from disk! Skipping.."
                + colorText.END
            )
        if defaultAnswer is None:
            OutputControls().takeUserInput(
                colorText.GREEN
                + "  [+] Press <Enter> to continue.."
                + colorText.END
            )

    def formattedBacktestOutput(outcome, pnlStats=False, htmlOutput=True):
        if not pnlStats:
            if htmlOutput:
                if outcome >= 80:
                    return f'{colorText.GREEN}{"%.2f%%" % outcome}{colorText.END}'
                if outcome >= 60:
                    return f'{colorText.WARN}{"%.2f%%" % outcome}{colorText.END}'
                return f'{colorText.FAIL}{"%.2f%%" % outcome}{colorText.END}'
            else:
                return f'{colorText.GREEN}{"%.2f%%" % outcome}{colorText.END}'
        else:
            if outcome >= 0:
                return f'{colorText.GREEN}{"%.2f%%" % outcome}{colorText.END}'
            return f'{colorText.FAIL}{"%.2f%%" % outcome}{colorText.END}'

    def getFormattedBacktestSummary(x, pnlStats=False, columnName=None):
        if x is not None and "%" in str(x):
            values = x.split("%")
            if (
                len(values) == 2
                and columnName is not None
                and ("-Pd" in columnName or "Overall" in columnName)
            ):
                return "{0}{1}".format(
                    PKConsoleTools.formattedBacktestOutput(
                        float(values[0]), pnlStats=pnlStats, htmlOutput=False
                    ),
                    values[1],
                )
        return x



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""

# import math
# import sys
# import warnings
# import datetime
# import numpy as np
# import os
# warnings.simplefilter("ignore", DeprecationWarning)
# warnings.simplefilter("ignore", FutureWarning)
# import pandas as pd

# from sys import float_info as sflt
# import pkscreener.classes.Utility as Utility
# from pkscreener import Imports
# from pkscreener.classes.Pktalib import pktalib
# from PKDevTools.classes.OutputControls import OutputControls
# from PKDevTools.classes import Archiver
# from PKNSETools.morningstartools import Stock

# if sys.version_info >= (3, 11):
#     import advanced_ta as ata

# # from sklearn.preprocessing import StandardScaler
# if Imports["scipy"]:
#     from scipy.stats import linregress

# from PKDevTools.classes.ColorText import colorText
# from PKDevTools.classes.PKDateUtilities import PKDateUtilities
# from PKDevTools.classes.SuppressOutput import SuppressOutput
# from PKDevTools.classes.MarketHours import MarketHours
# # from PKDevTools.classes.log import measure_time

# # Exception for only downloading stock data and not screening
# class DownloadDataOnly(Exception):
#     pass

# class EligibilityConditionNotMet(Exception):
#     pass

# # Exception for stocks which are not newly listed when screening only for Newly Listed
# class NotNewlyListed(Exception):
#     pass


# # Exception for stocks which are not stage two
# class NotAStageTwoStock(Exception):
#     pass

# # Exception for LTP not being in the range as per config
# class LTPNotInConfiguredRange(Exception):
#     pass

# # Exception for stocks which are low in volume as per configuration of 'minimumVolume'
# class NotEnoughVolumeAsPerConfig(Exception):
#     pass


# # Exception for newly listed stocks with candle nos < daysToLookback
# class StockDataNotAdequate(Exception):
#     pass

# class BaseScreeningStatistics:



#!/usr/bin/env python
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
# import os
# from time import sleep
# from pkscreener.classes import Utility, ImageUtility

# from PKDevTools.classes.log import default_logger
# from PKDevTools.classes.Telegram import (
#     is_token_telegram_configured,
#     send_document,
#     send_message,
#     send_photo,
#     send_media_group
# )

# DEV_CHANNEL_ID="-1001785195297"

# class PKMessenger:
    
#     @classmethod
#     def sendQuickScanResult(
#         self,
#         menuChoiceHierarchy,
#         user,
#         tabulated_results,
#         markdown_results,
#         caption,
#         pngName,
#         pngExtension,
#         addendum=None,
#         addendumLabel=None,
#         backtestSummary="",
#         backtestDetail="",
#         summaryLabel = None,
#         detailLabel = None,
#         legendPrefixText = "",
#         forceSend=False
#     ):
#         if "PKDevTools_Default_Log_Level" not in os.environ.keys():
#             if (("RUNNER" not in os.environ.keys()) or ("RUNNER" in os.environ.keys() and os.environ["RUNNER"] == "LOCAL_RUN_SCANNER")):
#                 return
#         try:
#             if not is_token_telegram_configured():
#                 return
#             ImageUtility.PKImageTools.tableToImage(
#                 markdown_results,
#                 tabulated_results,
#                 pngName + pngExtension,
#                 menuChoiceHierarchy,
#                 backtestSummary=backtestSummary,
#                 backtestDetail=backtestDetail,
#                 addendum=addendum,
#                 addendumLabel=addendumLabel,
#                 summaryLabel = summaryLabel,
#                 detailLabel = detailLabel,
#                 legendPrefixText = legendPrefixText
#             )
#             if forceSend:
#                 PKMessenger.sendMessageToTelegramChannel(
#                     message=None,
#                     document_filePath=pngName + pngExtension,
#                     caption=caption,
#                     user=user,
#                 )
#                 os.remove(pngName + pngExtension)
#         except Exception as e:  # pragma: no cover
#             default_logger().debug(e, exc_info=True)
#             pass

#     def sendMessageToTelegramChannel(
#         self,message=None, photo_filePath=None, document_filePath=None, caption=None, user=None, mediagroup=False,userPassedArgs=None, test_messages_queue=None, media_group_dict=None
#     ):
#         if ("RUNNER" not in os.environ.keys() and (userPassedArgs is not None and not userPassedArgs.log)) or (userPassedArgs is not None and userPassedArgs.telegram):
#             return
        
#         if user is None and userPassedArgs is not None and userPassedArgs.user is not None:
#             user = userPassedArgs.user
#         if not mediagroup:
#             if test_messages_queue is not None:
#                 test_messages_queue.append(f"message:{message}\ncaption:{caption}\nuser:{user}\ndocument:{document_filePath}")
#                 if len(test_messages_queue) >10:
#                     test_messages_queue.pop(0)
#             if user is not None and caption is not None:
#                 caption = f"{caption.replace('&','n')}."
#             if message is not None:
#                 try:
#                     message = message.replace("&", "n").replace("<","*")
#                     send_message(message, userID=user)
#                 except Exception as e:  # pragma: no cover
#                     default_logger().debug(e, exc_info=True)
#             else:
#                 message = ""
#             if photo_filePath is not None:
#                 try:
#                     if caption is not None:
#                         caption = f"{caption.replace('&','n')}"
#                     send_photo(photo_filePath, (caption if len(caption) <=1024 else ""), userID=user)
#                     # Breather for the telegram API to be able to send the heavy photo
#                     sleep(2)
#                 except Exception as e:  # pragma: no cover
#                     default_logger().debug(e, exc_info=True)
#             if document_filePath is not None:
#                 try:
#                     if caption is not None and isinstance(caption,str):
#                         caption = f"{caption.replace('&','n')}"
#                     send_document(document_filePath, (caption if len(caption) <=1024 else ""), userID=user)
#                     # Breather for the telegram API to be able to send the document
#                     sleep(2)
#                 except Exception as e:  # pragma: no cover
#                     default_logger().debug(e, exc_info=True)
#         else:
#             file_paths = []
#             file_captions = []
#             if "ATTACHMENTS" in media_group_dict.keys():
#                 attachments = media_group_dict["ATTACHMENTS"]
#                 numFiles = len(attachments)
#                 if numFiles >= 4:
#                     media_group_dict["ATTACHMENTS"] = []
#                 for attachment in attachments:
#                     file_paths.append(attachment["FILEPATH"])
#                     cleanCaption = attachment["CAPTION"].replace('&','n')[:1024]
#                     if "<pre>" in cleanCaption and "</pre>" not in cleanCaption:
#                         cleanCaption = f"{cleanCaption[:1018]}</pre>"
#                     file_captions.append(cleanCaption)
#                 if test_messages_queue is not None:
#                     test_messages_queue.append(f"message:{file_captions[-1]}\ncaption:{file_captions[-1]}\nuser:{user}\ndocument:{file_paths[-1]}")
#                     if len(test_messages_queue) >10:
#                         test_messages_queue.pop(0)
#                 if len(file_paths) > 0 and not userPassedArgs.monitor:
#                     resp = send_media_group(user=userPassedArgs.user,
#                                                     png_paths=[],
#                                                     png_album_caption=None,
#                                                     file_paths=file_paths,
#                                                     file_captions=file_captions)
#                     if resp is not None:
#                         default_logger().debug(resp.text, exc_info=True)
#                 caption = f"{str(len(file_captions))} files sent!"
#                 message = media_group_dict["CAPTION"].replace('&','n').replace("<","*")[:1024] if "CAPTION" in media_group_dict.keys() else "-"
#             for f in file_paths:
#                 try:
#                     if "RUNNER" in os.environ.keys():
#                         os.remove(f)
#                     elif not f.endswith("xlsx"):
#                         os.remove(f)
#                 except: # pragma: no cover
#                     pass
#         if user is not None:
#             if str(user) != str(DEV_CHANNEL_ID) and userPassedArgs is not None and not userPassedArgs.monitor:
#                 # Send an update to dev channel
#                 send_message(
#                     f"Responded back to userId:{user} with {caption}.{message} [{userPassedArgs.options.replace(':D','')}]",
#                     userID=DEV_CHANNEL_ID,
#                 )



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import os
import logging
import sys
import time
import warnings
import datetime
import numpy as np

warnings.simplefilter("ignore", DeprecationWarning)
warnings.simplefilter("ignore", FutureWarning)
import pandas as pd
# from PKDevTools.classes.log import tracelog
# from PKDevTools.classes.PKTimer import PKTimer
from PKDevTools.classes import Archiver, log
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.Fetcher import StockDataEmptyException
from PKDevTools.classes.SuppressOutput import SuppressOutput
from PKDevTools.classes.PKDateUtilities import PKDateUtilities

import pkscreener.classes.ScreeningStatistics as ScreeningStatistics
from pkscreener import Imports
from pkscreener.classes.CandlePatterns import CandlePatterns
from PKDevTools.classes.OutputControls import OutputControls

class StockScreener:
    def __init__(self):
        self.isTradingTime = PKDateUtilities.isTradingTime()
        self.configManager = None

    def setupLogger(self, log_level):
        if log_level > 0:
            os.environ["PKDevTools_Default_Log_Level"] = str(log_level)
        log.setup_custom_logger(
            "pkscreener",
            log_level,
            trace=False,
            log_file_path="pkscreener-logs.txt",
            filter=None,
        )

    # @tracelog
    def screenStocks(
        self,
        runOption,
        menuOption,
        exchangeName,
        executeOption,
        reversalOption,
        maLength,
        daysForLowestVolume,
        minRSI,
        maxRSI,
        respChartPattern,
        insideBarToLookback,
        totalSymbols,
        shouldCache,
        stock,
        newlyListedOnly,
        downloadOnly,
        volumeRatio,
        testbuild=False,
        userArgs=None,
        backtestDuration=0,
        backtestPeriodToLookback=30,
        logLevel=logging.NOTSET,
        portfolio=False,
        testData = None,
        hostRef=None,
    ):
        assert (
            hostRef is not None
        ), "hostRef argument must not be None. It should be an instance of PKMultiProcessorClient"
        if stock is None or len(stock) == 0:
            return None
        self.setupLogger(log_level=logLevel)
        configManager = hostRef.configManager
        self.configManager = configManager
        screeningDictionary, saveDictionary = self.initResultDictionaries()
        fullData = None
        processedData = None
        fetcher = hostRef.fetcher
        screener = hostRef.screener
        candlePatterns = hostRef.candlePatterns
        printCounter = userArgs.log if (userArgs is not None and userArgs.log is not None) else False
        userArgsLog = printCounter
        start_time = time.time()
        self.isTradingTime = False if menuOption in "B" else self.isTradingTime
        runOptionKey = runOption.split("=>")[0].split(":D:")[0].strip().replace(":0:",":12:")
        if menuOption in ["F"]:
            screeningDictionary["ScanOption"] = runOptionKey
            saveDictionary["ScanOption"] = runOptionKey
        # hostRef.default_logger.debug(f"runOption:{runOption}\nStock:{stock}\nmenuOption:{menuOption},\nexecuteOption:{executeOption},\nreversalOption:{reversalOption},\nmaLength:{maLength},\ndaysForLowestVolume:{daysForLowestVolume},\nminRSI:{minRSI},\nmaxRSI:{maxRSI},\nrespChartPattern:{respChartPattern},\ninsideBarToLookback:{insideBarToLookback}")
        # defaultsParentDict = {}
        # defaultsDict = {}
        # import json
        # with open("defaults.json","a+") as f:
        #     try:
        #         defaultsParentDict = json.loads(f.read())
        #     except: # pragma: no cover
        #         pass
        #     defaultsDict["reversalOption"] = reversalOption
        #     defaultsDict["maLength"] = maLength
        #     defaultsDict["daysForLowestVolume"] = daysForLowestVolume
        #     defaultsDict["minRSI"] = minRSI
        #     defaultsDict["maxRSI"] = maxRSI
        #     defaultsDict["respChartPattern"] = respChartPattern
        #     defaultsDict["insideBarToLookback"] = insideBarToLookback
        #     defaultsParentDict[runOption.split("=>")[0].replace(":SBIN,","").strip()] = defaultsDict
        #     allDdefaults = json.dumps(defaultsParentDict)
        #     f.write(allDdefaults)
        try:
            with hostRef.processingCounter.get_lock():
                hostRef.processingCounter.value += 1

            volumeRatio, period = self.determineBasicConfigs(stock, newlyListedOnly, volumeRatio, logLevel, hostRef, configManager, screener, userArgsLog)
            # if userArgsLog:
            #     hostRef.default_logger.info(f"For stock:{stock}, stock exists in objectDictionary:{hostRef.objectDictionaryPrimary.get(stock)}, cacheEnabled:{configManager.cacheEnabled}, isTradingTime:{self.isTradingTime}, downloadOnly:{downloadOnly}")
            data = None
            intraday_data = None
            data = self.getRelevantDataForStock(totalSymbols, shouldCache, stock, downloadOnly, printCounter, backtestDuration, hostRef,hostRef.objectDictionaryPrimary, configManager, fetcher, period,None, testData,exchangeName)
            if str(executeOption) in ["32","38","33"] or (not configManager.isIntradayConfig() and configManager.calculatersiintraday):
                # Daily data is already available in "data" above.
                # We need the intraday data for 1-d RSI values when config is not for intraday
                intraday_data = self.getRelevantDataForStock(totalSymbols, shouldCache, stock, downloadOnly, printCounter, backtestDuration, hostRef, hostRef.objectDictionarySecondary, configManager, fetcher, ("5d" if str(executeOption) in ["33"] else "1d"),"1m" if (str(executeOption) in ["33"] and maLength==3) else ("1m" if configManager.period.endswith("d") else configManager.duration), testData,exchangeName)
                
            if data is not None:
                if len(data) == 0 or data.empty or len(data) < backtestDuration:
                    raise StockDataEmptyException(f"Data length:{len(data)}")
            else:
                raise StockDataEmptyException(f"Data is None: {data}")
            
            bidGreaterThanAsk = False
            bidAskRatio = 0
            if executeOption == 29: # Bid vs Ask 
                hostRef.intradayNSEFetcher.symbol = stock.upper()
                priceData = hostRef.intradayNSEFetcher.price_order_info()
                if priceData is not None:
                    try:
                        totalBid = priceData["BidQty"].iloc[0]
                    except: # pragma: no cover
                        totalBid = 0
                        pass
                    try:
                        totalAsk = priceData["AskQty"].iloc[0]
                    except: # pragma: no cover
                        totalAsk = 0
                        pass
                    try:
                        lwrCP = float(priceData["LwrCP"].iloc[0])
                    except: # pragma: no cover
                        lwrCP = 0
                        pass
                    try:
                        uprCP = float(priceData["UprCP"].iloc[0])
                    except: # pragma: no cover
                        uprCP = 0
                        pass
                    try:
                        vwap = float(priceData["VWAP"].iloc[0])
                    except: # pragma: no cover
                        vwap = 0
                        pass
                    try:
                        dayVola = float(priceData["DayVola"].iloc[0])
                    except: # pragma: no cover
                        dayVola = 0
                        pass
                    try:
                        delPercent = priceData["Del(%)"].iloc[0]
                    except: # pragma: no cover
                        delPercent = 0
                        pass
                    try:
                        ltp = priceData["LTP"].iloc[0]
                    except: # pragma: no cover
                        ltp = 0
                        pass
                    
                    bidAskSimulate = userArgs is not None and userArgs.simulate is not None and "BidAsk" in userArgs.simulate.keys()
                    if (totalBid > totalAsk and \
                        ltp < uprCP and \
                        ltp > lwrCP) or bidAskSimulate:
                        bidGreaterThanAsk = True
                        bidAskRatio = round(totalBid/totalAsk,1) if totalAsk > 0 else (0 if not bidAskSimulate else 3)
                        bidAskBuildupDict = {"BidQty":totalBid,"AskQty":totalAsk,"LwrCP":lwrCP,"UprCP":uprCP,"VWAP":vwap,"DayVola":dayVola,"Del(%)":delPercent}
                        screeningDictionary.update(bidAskBuildupDict)
                        saveDictionary.update(bidAskBuildupDict)
                    else:
                        raise ScreeningStatistics.EligibilityConditionNotMet("Bid/Ask Eligibility Not met.")
                else:
                    raise ScreeningStatistics.EligibilityConditionNotMet("Bid/Ask Eligibility Not met.")
            # hostRef.default_logger.info(f"Will pre-process data:\n{data.tail(10)}")
            fullData, processedData, data = self.getCleanedDataForDuration(backtestDuration, portfolio, screeningDictionary, saveDictionary, configManager, screener, data)
            if "RUNNER" not in os.environ.keys() and backtestDuration == 0 and configManager.calculatersiintraday:
                if (intraday_data is not None and not intraday_data.empty):
                    intraday_fullData, intraday_processedData = screener.preprocessData(
                        intraday_data, daysToLookback=configManager.effectiveDaysToLookback
                    )
                    # Match the index length and values length
                    fullData = fullData.head(len(intraday_fullData))
                    intraday_fullData = intraday_fullData.head(len(fullData))
                    processedData = processedData.head(len(intraday_processedData))
                    intraday_processedData = intraday_processedData.head(len(processedData))
                    data = data.tail(len(intraday_data))
                    intraday_data = intraday_data.tail(len(data))
                    # Indexes won't match. Hence, we'd need to fallback on tolist
                    if "RSIi" not in processedData.columns:
                        processedData.insert(len(processedData.columns), "RSIi", intraday_processedData["RSI"].tolist())
                    if "RSIi" not in fullData.columns:
                        fullData.insert(len(fullData.columns), "RSIi", intraday_processedData["RSI"].tolist())
                else:
                    with SuppressOutput(suppress_stderr=(logLevel==logging.NOTSET), suppress_stdout=(not (printCounter or testbuild))):
                        if "RSIi" not in processedData.columns:
                            processedData.insert(len(processedData.columns), "RSIi", np.array(np.nan))
                            fullData.insert(len(fullData.columns), "RSIi", np.array(np.nan))
            else:
                    with SuppressOutput(suppress_stderr=(logLevel==logging.NOTSET), suppress_stdout=(not (printCounter or testbuild))):
                        if "RSIi" not in processedData.columns:
                            processedData.insert(len(processedData.columns), "RSIi", np.array(np.nan))
                            fullData.insert(len(fullData.columns), "RSIi", np.array(np.nan))

            def returnLegibleData(exceptionMessage=None):
                if backtestDuration == 0 or menuOption not in ["B"]:
                    raise ScreeningStatistics.EligibilityConditionNotMet(exceptionMessage)
                elif (backtestDuration > 0 and backtestDuration <= configManager.maxBacktestWindow):
                    screener.validateMovingAverages(
                        processedData, screeningDictionary, saveDictionary, maRange=1.25
                    )
                    screener.findTrend(
                        processedData,
                        screeningDictionary,
                        saveDictionary,
                        daysToLookback=configManager.daysToLookback,
                        stockName=stock,
                    )
                    screener.find52WeekHighLow(
                            fullData, saveDictionary, screeningDictionary
                        )
                    return (
                            screeningDictionary,
                            saveDictionary,
                            data,
                            stock,
                            backtestDuration,
                            runOptionKey
                        )
            if newlyListedOnly:
                if not screener.validateNewlyListed(fullData, period):
                    raise ScreeningStatistics.NotNewlyListed

            if processedData.empty:
                raise StockDataEmptyException("Empty processedData")
            suppressError = (logLevel==logging.NOTSET)
            suppressOut = (not (printCounter or testbuild))
            with SuppressOutput(suppress_stderr=suppressError, suppress_stdout=suppressOut):
                self.updateStock(stock, screeningDictionary, saveDictionary, executeOption, exchangeName,userArgs)
                
                self.performBasicLTPChecks(executeOption, screeningDictionary, saveDictionary, fullData, configManager, screener, exchangeName)
                hasMinVolumeRatio = self.performBasicVolumeChecks(executeOption, volumeRatio, screeningDictionary, saveDictionary, processedData, configManager, screener)
                if bidGreaterThanAsk:
                    if not hasMinVolumeRatio or bidAskRatio < 2:
                        raise ScreeningStatistics.EligibilityConditionNotMet("Bid/Ask Eligibility Not met.")
                isConfluence = False
                isInsideBar = 0
                isMaReversal = 0
                bullishCount = 0 
                bearishCount = 0
                isIpoBase = False
                isMaSupport = False
                isLorentzian = False
                isVCP = False
                isMinerviniVCP = False
                isVSA = False
                isNR = False
                hasPsarRSIReversal = False
                hasRisingRSIReversal = False
                hasRSIMAReversal = False
                isValidRsi = False
                isBuyingTrendline = False
                isMomentum = False
                mfiStake = 0
                fairValueDiff = 0
                consolidationValue = 0
                isBreaking = False
                isValidCci = False
                isVSA = False
                isCandlePattern = False
                isLowestVolume = False
                hasBbandsSqz = False
                hasMASignalFilter = False
                priceCrossed = False

                isValidityCheckMet = self.performValidityCheckForExecuteOptions(executeOption,screener,fullData,screeningDictionary,saveDictionary,processedData,configManager,maLength,intraday_data)
                if not isValidityCheckMet:
                    return returnLegibleData("Validity Check not met!")
                isShortTermBullish = (executeOption == 11 and isValidityCheckMet)
                if newlyListedOnly:
                    isIpoBase = screener.validateIpoBase(
                        stock, fullData, screeningDictionary, saveDictionary
                    )
                if executeOption in [1,2]:
                    isBreaking = screener.findBreakoutValue(
                        processedData,
                        screeningDictionary,
                        saveDictionary,
                        daysToLookback=configManager.daysToLookback,
                        alreadyBrokenout=(executeOption == 2),
                    )
                    if executeOption == 1:
                        isPotentialBreaking = screener.findPotentialBreakout(
                            fullData,
                            screeningDictionary,
                            saveDictionary,
                            daysToLookback=configManager.daysToLookback,
                        )
                        if not (isBreaking or isPotentialBreaking) or not hasMinVolumeRatio:
                            return returnLegibleData(f"isBreaking:{isBreaking},isPotentialBreaking:{isPotentialBreaking},hasMinVolumeRatio:{hasMinVolumeRatio}")
                    elif executeOption == 2:
                        if not (isBreaking) or not hasMinVolumeRatio:
                            return returnLegibleData(f"isBreaking:{isBreaking},hasMinVolumeRatio:{hasMinVolumeRatio}")
                elif executeOption == 3:
                    consolidationValue = screener.validateConsolidation(
                        processedData,
                        screeningDictionary,
                        saveDictionary,
                        percentage=configManager.consolidationPercentage,
                    )
                    if ((consolidationValue == 0 or consolidationValue > configManager.consolidationPercentage)):
                        return returnLegibleData(f"consolidationValue:{consolidationValue}")
                elif executeOption == 4:
                    isLowestVolume = screener.validateLowestVolume(
                        processedData, daysForLowestVolume
                    )
                    if not isLowestVolume:
                        return returnLegibleData(f"isLowestVolume:{isLowestVolume}")
                elif executeOption == 5:
                    isValidRsi = screener.validateRSI(
                        processedData, screeningDictionary, saveDictionary, minRSI, maxRSI
                    )
                    if not isValidRsi:
                        return returnLegibleData(f"isValidRsi:{isValidRsi}")
                elif executeOption == 6:
                    if reversalOption == 10:
                        hasRSIMAReversal = screener.findRSICrossingMA(processedData,
                                                                      screeningDictionary,
                                                                      saveDictionary,
                                                                      lookFor=maLength) # 1 =Buy, 2 =Sell, 3 = Any
                        if not hasRSIMAReversal:
                            return returnLegibleData(f"hasRSIMAReversal:{hasRSIMAReversal}")
                    elif reversalOption == 9:
                        hasRisingRSIReversal = screener.findRisingRSI(processedData)
                        if not hasRisingRSIReversal:
                            return returnLegibleData(f"hasRisingRSIReversal:{hasRisingRSIReversal}")
                    elif reversalOption == 8:
                        hasPsarRSIReversal = screener.findPSARReversalWithRSI(
                            processedData,
                            screeningDictionary,
                            saveDictionary
                            # minRSI=maLength if maLength is not None else 40,
                        )
                        if not hasPsarRSIReversal:
                            return returnLegibleData(f"hasPsarRSIReversal:{hasPsarRSIReversal}")
                    elif reversalOption == 6:
                        isNR = screener.validateNarrowRange(
                            processedData,
                            screeningDictionary,
                            saveDictionary,
                            nr=maLength if maLength is not None else 4,
                        )
                        if not isNR:
                            return returnLegibleData(f"isNR:{isNR}")
                    elif reversalOption == 5:
                        isVSA = screener.validateVolumeSpreadAnalysis(
                            processedData, screeningDictionary, saveDictionary
                        )
                        if not isVSA:
                            return returnLegibleData(f"isVSA:{isVSA}")
                    elif reversalOption == 4 and maLength is not None:
                        isMaSupport = screener.findReversalMA(
                            fullData, screeningDictionary, saveDictionary, maLength
                        )
                        if not isMaSupport:
                            return returnLegibleData(f"isMaSupport:{isMaSupport}")
                    elif reversalOption == 7:
                        if sys.version_info >= (3, 11):
                            isLorentzian = screener.validateLorentzian(
                                fullData,
                                screeningDictionary,
                                saveDictionary,
                                lookFor=maLength, # 1 =Buy, 2 =Sell, 3 = Any
                                stock=stock,
                            )
                            if not isLorentzian:
                                return returnLegibleData(f"isLorentzian:{isLorentzian}")
                elif executeOption == 7:
                    if respChartPattern == 3:
                        isConfluence = screener.validateConfluence(
                            stock,
                            processedData,
                            fullData,
                            screeningDictionary,
                            saveDictionary,
                            percentage=insideBarToLookback,
                            confFilter=(maLength if maLength > 0 else 3) # 1 = Conf up, 2 = Conf Down, 3 = all, 4 super confluence (10>20>55 EMA > 200SMA)
                        )
                        if not isConfluence:
                            return returnLegibleData(f"isConfluence:{isConfluence}")
                    elif respChartPattern == 4:
                        isVCP = screener.validateVCP(
                            fullData, screeningDictionary, saveDictionary,stockName=stock
                        )
                        if not isVCP:
                            return returnLegibleData(f"isVCP:{isVCP}")
                        else:
                            if hostRef.rs_strange_index > 0:
                                screener.findRSRating(index_rs_value=hostRef.rs_strange_index,df=fullData,screenDict=screeningDictionary, saveDict=saveDictionary)
                            screener.findRVM(df=fullData,screenDict=screeningDictionary, saveDict=saveDictionary)
                    elif respChartPattern == 5:
                        if Imports["scipy"]:
                            isBuyingTrendline = screener.findTrendlines(
                                fullData, screeningDictionary, saveDictionary
                            )
                            if not isBuyingTrendline:
                                return returnLegibleData(f"isBuyingTrendline:{isBuyingTrendline}")
                    elif respChartPattern == 6:
                        hasBbandsSqz = screener.findBbandsSqueeze(fullData, screeningDictionary, saveDictionary, filter=(maLength if maLength > 0 else 4))
                        if not hasBbandsSqz:
                            return returnLegibleData(f"hasBbandsSqz:{hasBbandsSqz}")
                    elif respChartPattern == 7:
                        try:
                            filterPattern = None
                            if str(maLength) != "0":
                                from pkscreener.classes.MenuOptions import CANDLESTICK_DICT
                                filterPattern = CANDLESTICK_DICT[str(maLength)]
                        except: # pragma: no cover
                            pass
                        if "Cup and Handle" in filterPattern:
                            isCandlePattern,_ = screener.find_cup_and_handle(fullData,saveDictionary,screeningDictionary,int(maLength))
                        else:
                            isCandlePattern = candlePatterns.findPattern(
                                processedData, screeningDictionary, saveDictionary,filterPattern)
                        if not isCandlePattern:
                            return returnLegibleData(f"isCandlePattern:{isCandlePattern}")
                    elif respChartPattern == 8:
                        isMinerviniVCP = screener.validateVCPMarkMinervini(
                            fullData, screeningDictionary, saveDictionary
                        )
                        if not isMinerviniVCP:
                            return returnLegibleData(f"isMinerviniVCP:{isMinerviniVCP}")
                        else:
                            if hostRef.rs_strange_index > 0:
                                screener.findRSRating(index_rs_value=hostRef.rs_strange_index,df=fullData,screenDict=screeningDictionary, saveDict=saveDictionary)
                            screener.findRVM(df=fullData,screenDict=screeningDictionary, saveDict=saveDictionary)
                    elif respChartPattern == 9:
                        hasMASignalFilter,_, _ = screener.validateMovingAverages(
                            fullData, screeningDictionary, saveDictionary,maRange=1.25,maLength=maLength
                        )
                        if not hasMASignalFilter:
                            return returnLegibleData(f"hasMASignalFilter:{hasMASignalFilter}")
                elif executeOption == 10:
                    isPriceRisingByAtLeast2Percent = (
                        screener.validatePriceRisingByAtLeast2Percent(
                            processedData, screeningDictionary, saveDictionary
                        )
                    )
                    if not isPriceRisingByAtLeast2Percent:
                        return returnLegibleData(f"isPriceRisingByAtLeast2Percent:{isPriceRisingByAtLeast2Percent}")
                # Must-run, but only at the end
                try:
                    if executeOption != 7 or (executeOption == 7 and respChartPattern != 7):
                    # Only 'doji' and 'inside' is internally implemented by pandas_ta_classic.
                    # Otherwise, for the rest of the candle patterns, they also need
                    # TA-Lib. So if TA-Lib is not available, it will throw exception
                    # We can live with no-patterns if user has not installed ta-lib
                    # yet. If ta-lib is available, PKTalib will load it automatically.
                        isCandlePattern = candlePatterns.findPattern(
                            processedData, screeningDictionary, saveDictionary
                        )
                except KeyboardInterrupt: # pragma: no cover
                    raise KeyboardInterrupt
                except Exception as e:  # pragma: no cover
                    hostRef.default_logger.debug(e, exc_info=True)
                    screeningDictionary["Pattern"] = ""
                    saveDictionary["Pattern"] = ""

                try:
                    currentTrend = screener.findTrend(
                        processedData,
                        screeningDictionary,
                        saveDictionary,
                        daysToLookback=configManager.daysToLookback,
                        stockName=stock,
                    )
                    if backtestDuration == 0:
                        if executeOption == 21 and reversalOption in [3,5,6,7,8,9]:
                            # Find general trend
                            _,mfiStake,fairValueDiff = screener.findUptrend(
                                fullData,
                                screeningDictionary,
                                saveDictionary,
                                testbuild,
                                stock,
                                onlyMF=(executeOption == 21 and reversalOption in [5,6]),
                                hostData=data,
                                exchangeName=exchangeName,
                                refreshMFAndFV=(menuOption in ["X", "C", "F"]),
                                downloadOnly=True
                            )
                            hostRef.objectDictionaryPrimary[stock] = data.to_dict("split")
                except np.RankWarning as e: # pragma: no cover 
                    hostRef.default_logger.debug(e, exc_info=True)
                    screeningDictionary["Trend"] = "Unknown"
                    saveDictionary["Trend"] = "Unknown"
                # CCI also uses "Trend" value from findTrend above.
                # So it must only be called after findTrend
                if executeOption == 8:
                    isValidCci = screener.validateCCI(
                        processedData, screeningDictionary, saveDictionary, minRSI, maxRSI
                    )
                    if not isValidCci:
                        return returnLegibleData(f"isValidCci:{isValidCci}")

                if not (isConfluence or isShortTermBullish or hasMASignalFilter):
                    isMaReversal,bullishCount, bearishCount = screener.validateMovingAverages(
                        processedData, screeningDictionary, saveDictionary, maRange=1.25
                    )
                if executeOption == 6:
                    if reversalOption == 1 and not (str(saveDictionary["Pattern"]).split(",")[0]
                                                                    in CandlePatterns.reversalPatternsBullish
                                                                    or isMaReversal > 0):
                        return returnLegibleData(f"reversalOption:{reversalOption},isMaReversal:{isMaReversal},{CandlePatterns.reversalPatternsBullish}")
                    elif reversalOption == 2 and not (str(saveDictionary["Pattern"]).split(",")[0]
                                                                    in CandlePatterns.reversalPatternsBearish
                                                                    or isMaReversal < 0):
                        return returnLegibleData(f"reversalOption:{reversalOption},isMaReversal:{isMaReversal},{CandlePatterns.reversalPatternsBearish}")
                # validateInsideBar needs "Trend" to be already defined
                # ValidateInsideBar also needs "MA-Signal" to be setup
                if executeOption == 7 and respChartPattern < 3:
                    isInsideBar = screener.validateInsideBar(
                                processedData,
                                screeningDictionary,
                                saveDictionary,
                                chartPattern=respChartPattern,
                                daysToLookback=insideBarToLookback,
                            )
                    if isInsideBar ==0:
                        return returnLegibleData(f"isInsideBar:{isInsideBar}")
                if executeOption == 40:
                    priceCrossed = screener.validatePriceActionCrosses(full_df=fullData,
                                                                  screenDict=screeningDictionary,
                                                                  saveDict=saveDictionary,
                                                                  mas=insideBarToLookback,
                                                                  isEMA=respChartPattern,
                                                                  maDirectionFromBelow=reversalOption)
                if executeOption == 41:
                    priceCrossed = screener.validatePriceActionCrossesForPivotPoint(df=processedData.head(2),
                                                                  screenDict=screeningDictionary,
                                                                  saveDict=saveDictionary,
                                                                  pivotPoint=respChartPattern,
                                                                  crossDirectionFromBelow=reversalOption)
                if not (isLorentzian or (isInsideBar !=0) or isBuyingTrendline or isIpoBase or isNR or isVCP or isVSA or isMinerviniVCP):
                    isMomentum = screener.validateMomentum(
                        processedData, screeningDictionary, saveDictionary
                    )
                    if executeOption == 6 and reversalOption ==3 and not isMomentum:
                        return returnLegibleData(f"executeOption:{executeOption},reversalOption:{reversalOption},isMomentum:{isMomentum}")

                with hostRef.processingResultsCounter.get_lock():
                    # hostRef.default_logger.debug(f"ExecuteOption:{executeOption}:{reversalOption}:{respChartPattern}:{maLength}. Elapsed: {time.time() - start_time}")
                    if (
                        (executeOption == 0)
                        or ((
                            (
                                (executeOption == 1 and (isBreaking or isPotentialBreaking))
                                or (executeOption == 2 and isBreaking)
                            )
                            and hasMinVolumeRatio
                        ))
                        or ((
                            (executeOption == 3)
                            and (
                                consolidationValue <= configManager.consolidationPercentage
                                and consolidationValue != 0
                            )
                        ))
                        or (executeOption == 4 and isLowestVolume)
                        or (executeOption == 5 and isValidRsi)
                        or ((executeOption == 6) and ((reversalOption == 1 and (
                                                                    str(saveDictionary["Pattern"]).split(",")[0]
                                                                    in CandlePatterns.reversalPatternsBullish
                                                                    or isMaReversal > 0
                                                                ))
                                                                or (reversalOption == 2 and (
                                                                    str(saveDictionary["Pattern"]).split(",")[0]
                                                                    in CandlePatterns.reversalPatternsBearish
                                                                    or isMaReversal < 0
                                                                ))
                                                                or (reversalOption == 3 and isMomentum)
                                                                or (reversalOption == 4 and isMaSupport)
                                                                or ((
                                                                    reversalOption == 5
                                                                    and isVSA
                                                                    and saveDictionary["Pattern"]
                                                                    in CandlePatterns.reversalPatternsBullish
                                                                ))
                                                                or (reversalOption == 6 and isNR)
                                                                or (reversalOption == 7 and isLorentzian)
                                                                or (reversalOption == 8 and hasPsarRSIReversal)
                                                                or (reversalOption == 9 and hasRisingRSIReversal)
                                                                or (reversalOption == 10 and hasRSIMAReversal)
                                                                ))
                        or ((executeOption == 7) and ((respChartPattern < 3 and isInsideBar > 0) 
                                                                  or (isConfluence)
                                                                  or (isIpoBase and newlyListedOnly and not respChartPattern < 3)
                                                                  or (isVCP and ((bearishCount == 0) if configManager.enableAdditionalVCPEMAFilters else True))
                                                                  or (isBuyingTrendline)
                                                                  or (respChartPattern == 6 and hasBbandsSqz)
                                                                  or (respChartPattern == 7 and isCandlePattern)
                                                                  or (respChartPattern == 8 and isMinerviniVCP)
                                                                  or (respChartPattern == 9 and hasMASignalFilter)))
                        or (executeOption == 8 and isValidCci)
                        or (executeOption == 9 and hasMinVolumeRatio)
                        or (executeOption == 10 and isPriceRisingByAtLeast2Percent)
                        or (executeOption == 11 and isShortTermBullish)
                        or (executeOption in [12,13,14,15,16,17,18,19,20,23,24,25,27,28,30,31,32,33,34,35,36,37,38,39,42,43] and isValidityCheckMet)
                        or (executeOption == 21 and (mfiStake > 0 and reversalOption in [3,5]))
                        or (executeOption == 21 and (mfiStake < 0 and reversalOption in [6,7]))
                        or (executeOption == 21 and (fairValueDiff > 0 and reversalOption in [8]))
                        or (executeOption == 21 and (fairValueDiff < 0 and reversalOption in [9]))
                        or (executeOption == 26)
                        or (executeOption == 29 and bidGreaterThanAsk)
                        or (executeOption == 40 and priceCrossed)
                        or (executeOption == 41 and priceCrossed)
                    ):
                        isNotMonitoringDashboard = userArgs is None or userArgs.monitor is None or (userArgs.monitor is not None and "~" not in userArgs.monitor)
                        # Now screen for common ones to improve performance
                        if isNotMonitoringDashboard and not (executeOption == 6 and reversalOption == 7):
                            if sys.version_info >= (3, 11):
                                with SuppressOutput(suppress_stderr=True, suppress_stdout=True):
                                    screener.validateLorentzian(
                                        fullData,
                                        screeningDictionary,
                                        saveDictionary,
                                        lookFor=maLength, # 1 =Buy, 2 =Sell, 3 = Any
                                        stock=stock,
                                    )
                        if isNotMonitoringDashboard and not (executeOption in [1,2]):
                            screener.findBreakoutValue(
                                processedData,
                                screeningDictionary,
                                saveDictionary,
                                daysToLookback=configManager.daysToLookback,
                                alreadyBrokenout=(executeOption == 2),
                            )
                        if (isNotMonitoringDashboard and executeOption != 3) or (self.configManager.alwaysExportToExcel):
                            screener.validateConsolidation(
                                processedData,
                                screeningDictionary,
                                saveDictionary,
                                percentage=configManager.consolidationPercentage,
                            )
                        if executeOption != 5:
                            screener.validateRSI(
                                processedData, screeningDictionary, saveDictionary, minRSI, maxRSI
                            )
                        screener.find52WeekHighLow(
                            fullData, saveDictionary, screeningDictionary
                        )
                        if isNotMonitoringDashboard and executeOption != 8:
                            screener.validateCCI(
                                processedData, screeningDictionary, saveDictionary, minRSI, maxRSI
                            )
                        if configManager.enableAdditionalTrendFilters and isNotMonitoringDashboard and executeOption != 21 and backtestDuration == 0:
                            # We don't need to have MFI or fair value data for backtesting because those
                            # are anyways only available for days in the past.
                            # For executeOption 21, we'd have already got the mfiStake and fairValueDiff
                            # Find general trend, MFI data and fairvalue only after the stocks are already screened
                            screener.findUptrend(
                                fullData,
                                screeningDictionary,
                                saveDictionary,
                                testbuild,
                                stock,
                                onlyMF=(executeOption == 21 and reversalOption in [5,6]),
                                hostData=data,
                                exchangeName=exchangeName,
                                downloadOnly=downloadOnly
                            )
                            hostRef.objectDictionaryPrimary[stock] = data.to_dict("split")
                        if userArgs is not None and userArgs.usertag is not None and "VCP" in userArgs.usertag:
                            if hostRef.rs_strange_index > 0:
                                if f"RS_Rating{self.configManager.baseIndex}" not in saveDictionary.keys():
                                    screener.findRSRating(index_rs_value=hostRef.rs_strange_index,df=fullData,screenDict=screeningDictionary, saveDict=saveDictionary)
                            if "RVM" not in saveDictionary.keys():
                                screener.findRVM(df=fullData,screenDict=screeningDictionary, saveDict=saveDictionary)
                        hostRef.processingResultsCounter.value += 1
                        return (
                            screeningDictionary,
                            saveDictionary,
                            data,
                            stock,
                            backtestDuration,
                            runOptionKey
                        )

        except KeyboardInterrupt: # pragma: no cover
            # Capturing Ctr+C Here isn't a great idea
            pass
        except StockDataEmptyException as e: # pragma: no cover
            if data is None or (data is not None and not data.isnull().values.all(axis=0)[0]):
                hostRef.default_logger.debug(f"StockDataEmptyException:{stock}: {e}", exc_info=True)
            pass
        except ScreeningStatistics.EligibilityConditionNotMet as e: # pragma: no cover
            if userArgsLog:
                hostRef.default_logger.debug(f"EligibilityConditionNotMet:{stock}: {e}", exc_info=True)
            pass
        except ScreeningStatistics.NotNewlyListed as e: # pragma: no cover
            if userArgsLog:
                hostRef.default_logger.debug(f"NotNewlyListed:{stock}: {e}", exc_info=True)
            pass
        except ScreeningStatistics.NotAStageTwoStock as e: # pragma: no cover
            if userArgsLog:
                hostRef.default_logger.debug(f"NotAStageTwoStock:{stock}: {e}", exc_info=True)
            pass
        except ScreeningStatistics.NotEnoughVolumeAsPerConfig as e: # pragma: no cover 
            if userArgsLog:
                hostRef.default_logger.debug(f"NotEnoughVolumeAsPerConfig:{stock}: {e}", exc_info=True)
            pass
        except ScreeningStatistics.DownloadDataOnly as e: # pragma: no cover
            if userArgsLog:
                hostRef.default_logger.debug(f"DownloadDataOnly:{stock}: {e}", exc_info=True)
            try:
                data = hostRef.objectDictionaryPrimary.get(stock)
                if data is not None:
                    data = pd.DataFrame(data["data"], columns=data["columns"], index=data["index"])
                    screener.getMutualFundStatus(stock, hostData=data, force=True, exchangeName=exchangeName)
                    hostRef.objectDictionaryPrimary[stock] = data.to_dict("split")
            except KeyboardInterrupt: # pragma: no cover
                raise KeyboardInterrupt
            except Exception as ex:
                hostRef.default_logger.debug(f"MFIStatus: {stock}:\n{ex}", exc_info=True)
                pass
            try:
                screener.getFairValue(stock,hostData=data, force=True,exchangeName=exchangeName)
                hostRef.objectDictionaryPrimary[stock] = data.to_dict("split")
            except KeyboardInterrupt: # pragma: no cover
                raise KeyboardInterrupt
            except Exception as ex:
                hostRef.default_logger.debug(f"FairValue: {stock}:\n{ex}", exc_info=True)
                pass
            pass
        except ScreeningStatistics.LTPNotInConfiguredRange as e: # pragma: no cover
            if userArgsLog:
                hostRef.default_logger.debug(f"LTPNotInConfiguredRange:{stock}: {e}", exc_info=True)
            pass
        except KeyError as e: # pragma: no cover
            if userArgsLog:
                hostRef.default_logger.debug(f"KeyError:{stock}: {e}", exc_info=True)
            pass
        except OSError as e: # pragma: no cover
            if userArgsLog:
                hostRef.default_logger.debug(f"OSError:{stock}: {e}", exc_info=True)
            pass
        except Exception as e:  # pragma: no cover
            if userArgsLog:
                hostRef.default_logger.debug(f"Exception:{stock}: {e}", exc_info=True)
            if testbuild or printCounter:
                import traceback
                traceback.print_exc()
                OutputControls().printOutput(e)
                OutputControls().printOutput(
                    colorText.FAIL
                    + (
                        "\n  [+] Exception Occured while Screening %s! Skipping this stock.."
                        % stock
                    )
                    + colorText.END
                )
        return None

    def performValidityCheckForExecuteOptions(self,executeOption,screener,fullData,screeningDictionary,saveDictionary,processedData,configManager,subMenuOption=3,intraday_data=None):
        isValid = True
        if executeOption not in [11,12,13,14,15,16,17,18,19,20,23,24,25,27,28,30,31,32,33,34,35,36,37,38,39,42,43]:
            return True
        if executeOption == 11:
            isValid = screener.validateShortTermBullish(
                fullData, screeningDictionary, saveDictionary
            )
        elif executeOption == 12:
            isValid = (
                screener.validate15MinutePriceVolumeBreakout(fullData)
            )
        elif executeOption == 13:
            isValid = screener.findBullishIntradayRSIMACD(
                fullData
            )
        elif executeOption == 14:
            isValid = screener.findNR4Day(fullData)
        elif executeOption == 15:
            isValid = screener.find52WeekLowBreakout(fullData)
        elif executeOption == 16:
            isValid = screener.find10DaysLowBreakout(fullData)
        elif executeOption == 17:
            isValid = screener.find52WeekHighBreakout(fullData)
        elif executeOption == 18:
            isValid = screener.findAroonBullishCrossover(fullData)
        elif executeOption == 19:
            isValid = screener.validateMACDHistogramBelow0(fullData)
        elif executeOption == 20:
            isValid = screener.validateBullishForTomorrow(fullData)
        elif executeOption == 23:
            isValid = screener.findBreakingoutNow(processedData, fullData, saveDictionary, screeningDictionary)
        elif executeOption == 24:
            isValid = (
                screener.validateHigherHighsHigherLowsHigherClose(fullData)
            )
        elif executeOption == 25:
            isValid = screener.validateLowerHighsLowerLows(processedData)
        elif executeOption == 27:
            isValid = screener.findATRCross(processedData,saveDictionary, screeningDictionary)
        elif executeOption == 28:
            isValid = screener.findHigherBullishOpens(processedData)
        elif executeOption == 30: # findBuySellSignalsFromATRTrailing # findATRTrailingStops
            isValid = screener.findATRTrailingStops(fullData,sensitivity=configManager.atrTrailingStopSensitivity, atr_period=configManager.atrTrailingStopPeriod,ema_period=configManager.atrTrailingStopEMAPeriod,buySellAll=subMenuOption,saveDict=saveDictionary,screenDict=screeningDictionary)
        elif executeOption == 31: # findBuySellSignalsFromATRTrailing # findATRTrailingStops
            isValid = screener.findHighMomentum(processedData,strict=(subMenuOption==1))
        elif executeOption == 32: # findIntradayOpenSetup
            isValid = screener.findIntradayOpenSetup(processedData,intraday_data,saveDictionary,screeningDictionary,buySellAll=subMenuOption)
        elif executeOption == 33:
            if subMenuOption == 1:
                isValid = screener.findPotentialProfitableEntriesFrequentHighsBullishMAs(processedData,fullData, saveDictionary,screeningDictionary)
            elif subMenuOption == 2:
                isValid = screener.findPotentialProfitableEntriesBullishTodayForPDOPDC(processedData,saveDictionary,screeningDictionary)
            elif subMenuOption == 3:
                isValid = screener.findPotentialProfitableEntriesForFnOTradesAbove50MAAbove200MA5Min(intraday_data,fullData,saveDictionary,screeningDictionary)
        elif executeOption == 34: # findBullishAVWAP
            isValid = screener.findBullishAVWAP(fullData,screeningDictionary,saveDictionary)
        elif executeOption == 35: # findPerfectShortSellsFutures
            isValid = screener.findPerfectShortSellsFutures(fullData)
        elif executeOption == 36: # findProbableShortSellsFutures
            isValid = screener.findProbableShortSellsFutures(fullData)
        elif executeOption == 37: # findShortSellCandidatesForVolumeSMA
            isValid = screener.findShortSellCandidatesForVolumeSMA(fullData)
        elif executeOption == 38: # findIntradayShortSellWithPSARVolumeSMA
            isValid = screener.findIntradayShortSellWithPSARVolumeSMA(fullData,intraday_data)
        elif executeOption == 39: # findIPOLifetimeFirstDayBullishBreak
            isValid = screener.findIPOLifetimeFirstDayBullishBreak(fullData)
        elif executeOption == 42:
            isValid = screener.findSuperGainersLosers(fullData,subMenuOption)
        elif executeOption == 43:
            isValid = screener.findSuperGainersLosers(fullData,subMenuOption,gainer=False)
        return isValid        
                    
    def performBasicVolumeChecks(self, executeOption, volumeRatio, screeningDictionary, saveDictionary, processedData, configManager, screener):
        minVolume = configManager.minVolume / (
                    100 if configManager.isIntradayConfig() else 1
                )
        hasMinVolumeRatio, hasMinVolQty = screener.validateVolume(
                    processedData,
                    screeningDictionary,
                    saveDictionary,
                    volumeRatio=volumeRatio,
                    minVolume=minVolume,
                )
        if (not hasMinVolQty and executeOption > 0) or (executeOption == 9 and not hasMinVolumeRatio):
            raise ScreeningStatistics.NotEnoughVolumeAsPerConfig(f"hasMinVolQty:{hasMinVolQty},executeOption:{executeOption},hasMinVolumeRatio:{hasMinVolumeRatio}")
        return hasMinVolumeRatio

    def performBasicLTPChecks(self, executeOption, screeningDictionary, saveDictionary, fullData, configManager, screener,exchangeName):
        isLtpValid, verifyStageTwo = screener.validateLTP(
                    fullData,
                    screeningDictionary,
                    saveDictionary,
                    minLTP=configManager.minLTP if exchangeName == "INDIA" else configManager.minLTP/80,
                    maxLTP=configManager.maxLTP,
                    minChange=configManager.minimumChangePercentage
                )
        if not isLtpValid:
            raise ScreeningStatistics.LTPNotInConfiguredRange
        if configManager.stageTwo and not verifyStageTwo and (executeOption > 0 and executeOption not in [29]):
            raise ScreeningStatistics.NotAStageTwoStock

    def updateStock(self, stock, screeningDictionary, saveDictionary, executeOption=0,exchangeName='INDIA',userArgs=None):
        doNotAnchorText = executeOption == 26 or (userArgs is not None and userArgs.systemlaunched)
        screeningDictionary["Stock"] = (
                    colorText.WHITE
                    + (f"\x1B]8;;https://in.tradingview.com/chart?symbol={'NSE' if exchangeName=='INDIA' else 'NASDAQ'}%3A{stock}\x1B\\{stock}\x1B]8;;\x1B\\")
                    + colorText.END
                ) if not doNotAnchorText else stock
        saveDictionary["Stock"] = stock

    def getCleanedDataForDuration(self, backtestDuration, portfolio, screeningDictionary, saveDictionary, configManager, screener, data):
        fullData = None
        processedData = None
        ohlc_dict = {
            "open":'first',
            "high":'max',
            "low":'min',
            "close":'last',
            'Adj Close': 'last',
            "volume":'sum'
        }
        candleDuration = self.configManager.candleDurationInt
        candleDurationFrequency = self.configManager.candleDurationFrequency
        durationFrequency = "T" if candleDurationFrequency=="m" else ("H" if candleDurationFrequency=="h" else ("M" if candleDurationFrequency=="mo" else ("W" if candleDurationFrequency=="wk" else "T")))
        if int(candleDuration) >= 1 and (candleDurationFrequency in ["m","h","mo","wk"]):
            data = data.resample(f'{candleDuration}{durationFrequency}', offset='15min').agg(ohlc_dict)
            data = data[data["high"]>0] # resampling can introduce 0 value rows for non-market hours
        if backtestDuration == 0:
            fullData, processedData = screener.preprocessData(
                    data, daysToLookback=configManager.effectiveDaysToLookback
                )
            if processedData.empty:
                raise StockDataEmptyException(f"Empty processedData with data length ({len(data)})")
            if portfolio:
                data = data[::-1]
                
                screener.validateLTPForPortfolioCalc(
                        data, screeningDictionary, saveDictionary,requestedPeriod=backtestDuration
                    )
                data = data[::-1]
        else:
            if data is None or fullData is None or processedData is None:
                    # data will have the oldest date at the top and the most recent
                    # date will be at the bottom
                    # We want to have the nth day treated as today when pre-processing where n = backtestDuration row from the bottom
                inputData = data.head(len(data) - backtestDuration)
                    # imputData will have the last row as the date for which the entire calculation
                    # and prediction is being done
                data = data.tail(
                        backtestDuration + 1
                    )  # .head(backtestPeriodToLookback+1)
                    # Let's get today's data
                if portfolio:
                    screener.validateLTPForPortfolioCalc(
                            data, screeningDictionary, saveDictionary,requestedPeriod=backtestDuration
                        )
                    # data has the last row from inputData at the top.
                fullData, processedData = screener.preprocessData(
                        inputData, daysToLookback=configManager.daysToLookback
                    )
                
        return fullData,processedData,data

    def getRelevantDataForStock(self, totalSymbols, shouldCache, stock, downloadOnly, printCounter, backtestDuration, hostRef,objectDictionary, configManager, fetcher, period, duration, testData=None,exchangeName="INDIA"):
        hostData = objectDictionary.get(stock) if (objectDictionary is not None and len(objectDictionary) > 0) else None
        data = None
        hostDataLength = 0 if hostData is None else (0 if "data" not in hostData.keys() else len(hostData["data"]))
        start = None
        lastTradingDate = PKDateUtilities.tradingDate().strftime("%Y-%m-%d")
        if (configManager.candlePeriodFrequency in ["d","mo"] and configManager.candleDurationFrequency in ["m","h"]):
            if backtestDuration > 0: # We are backtesting
                start = PKDateUtilities.nthPastTradingDateStringFromFutureDate(backtestDuration)
                end = PKDateUtilities.nthPastTradingDateStringFromFutureDate(backtestDuration-1)
            else:
                # Since this is intraday data, we'd just need to start from the last trading session
                start = lastTradingDate
                end = PKDateUtilities.currentDateTime().strftime("%Y-%m-%d")
        if (
                not shouldCache
                or (downloadOnly and hostData is None)
                or (hostData is None and self.isTradingTime)
                or hostData is None or hostDataLength == 0
            ):
            if testData is not None:
                data = testData
                data.reset_index(inplace=True)
                data["Date"] = pd.to_datetime(data['index'].astype(int)/1000,unit='s')
                data.set_index("Date", inplace=True)
                data.drop("index",axis=1,inplace=True)
            else:
                # The following code tries to re-fetch the missing stock data
                # It has been commented because it's less likely to fetch the data again
                # within seconds of the first try when we must have tried to fetch data
                # in batches.
                # data = fetcher.fetchStockData(
                #         stock,
                #         period,
                #         configManager.duration if duration is None else duration,
                #         hostRef.proxyServer,
                #         hostRef.processingResultsCounter,
                #         hostRef.processingCounter,
                #         totalSymbols,
                #         start=start,
                #         end=start,
                #         exchangeSuffix=".NS" if exchangeName == "INDIA" else "",
                #         printCounter=printCounter
                #     )
                if hostData is not None and data is not None:
                    # During the market trading hours, we don't want to go for MFI/FV value fetching
                    # So let's copy the old saved ones.
                    try:
                        data["MF"] = hostData["MF"]
                        data["MF_Date"] = hostData["MF_Date"]
                    except KeyError:
                        pass
                    try:
                        data["FII"] = hostData["FII"]
                        data["FII_Date"] = hostData["FII_Date"]
                    except KeyError:
                        pass
                    try:
                        data["FairValue"] = hostData["FairValue"]
                    except KeyError:
                        pass
        else:
            self.printProcessingCounter(totalSymbols, stock, printCounter, hostRef)
            # data = hostData
            try:
                columns = hostData["columns"]
                data = pd.DataFrame(
                        hostData["data"], columns=columns, index=hostData["index"]
                    )
            except (ValueError, AssertionError) as e: # pragma: no cover
                # 9 columns passed, passed data had 11 columns
                # 10 columns passed, passed data had 11 columns
                excLookingFor = " columns passed, passed data had "
                if excLookingFor in str(e):
                    e_diff = str(e).replace(excLookingFor,",").replace(" columns","").split(",")
                    num_diff = int(e_diff[1]) - int(e_diff[0])
                    while (num_diff > 0):
                        columns.append(f"temp{num_diff}")
                        num_diff -= 1
                    data = pd.DataFrame(
                            hostData["data"], columns=columns, index=hostData["index"]
                        )
                else:
                    hostRef.default_logger.debug(e, exc_info=True)
                pass
        if "Datetime" in data.columns: # for intraday data, the column name is Datetime
            with pd.option_context('mode.chained_assignment', None):
                data["Date"] = data["Datetime"]
        try:
            data.reset_index(inplace=True)
            if "Datetime" in data.columns and "Date" not in data.columns:
                data.rename(columns={"Datetime": "Date"}, inplace=True)
            else:
                data.rename(columns={"index": "Date"}, inplace=True)
            data.set_index("Date", inplace=True)
            data.index = pd.to_datetime(data.index)
        except: # pragma: no cover
            pass
        if ((shouldCache and not self.isTradingTime and (hostData is None  or hostDataLength == 0)) or downloadOnly) \
            or (shouldCache and hostData is None):  # and backtestDuration == 0 # save only if we're NOT backtesting
                if start is None or start is lastTradingDate and data is not None:
                    objectDictionary[stock] = data.to_dict("split")
                if downloadOnly:
                    with hostRef.processingResultsCounter.get_lock():
                        hostRef.processingResultsCounter.value += 1
                    raise ScreeningStatistics.DownloadDataOnly
                else:
                    hostData = objectDictionary.get(stock)
        return data

    def determineBasicConfigs(self, stock, newlyListedOnly, volumeRatio, logLevel, hostRef, configManager, screener, userArgsLog):
        if userArgsLog:
            self.setupLoggers(hostRef, screener, logLevel, stock, userArgsLog=True)
        period = configManager.period
        if volumeRatio <= 0:
            volumeRatio = configManager.volumeRatio
            # Data download adjustment for Newly Listed only feature
        if newlyListedOnly:
            if configManager.period.endswith("y") and int(configManager.period[:-1]) >= 1:
                period = "220d"
            elif configManager.period.endswith("d") and int(configManager.period[:-1]) >= 250:
                period = "220d"
            else:
                period = configManager.period
        return volumeRatio,period

    def printProcessingCounter(self, totalSymbols, stock, printCounter, hostRef):
        if printCounter:
            try:
                OutputControls().printOutput(
                            colorText.GREEN
                            + (
                                "[%d%%] Screened %d, Found %d. Fetching data & Analyzing %s..."
                                % (
                                    int(
                                        (hostRef.processingCounter.value / totalSymbols)
                                        * 100
                                    ),
                                    hostRef.processingCounter.value,
                                    hostRef.processingResultsCounter.value,
                                    stock,
                                )
                            )
                            + colorText.END,
                            end="",
                        )
                OutputControls().printOutput(
                            colorText.GREEN
                            + "=> Done!"
                            + colorText.END,
                            end="\r",
                            flush=True,
                        )
            except ZeroDivisionError as e: # pragma: no cover
                hostRef.default_logger.debug(e, exc_info=True)
                pass
            sys.stdout.write("\r\033[K")
    
    def setupLoggers(self, hostRef, screener, logLevel, stock, userArgsLog=False):
        # Set the loglevels for both the caller and screener
        # Also add handlers that are specific to this sub-process which
        # will continue with the screening. Each sub-process would have
        # its own logger but going into the same file/console > to that
        # of the parent logger.
        screener.default_logger = hostRef.default_logger
        screener.shouldLog = userArgsLog

    def initResultDictionaries(self):
        periods = self.configManager.periodsRange
        columns = [
            "Stock",
            "LTP",
            "%Chng",
            "52Wk-H",
            "52Wk-L",
            "RSI",
            "RSIi",
            "volume",
            "22-Pd",
            "Consol.",
            "Breakout",
            "MA-Signal",
            "Trend",
            "Pattern",
            "CCI",
            "FairValue"
        ]
        screeningDictionary = {
            "Stock": "",
            "LTP": 0,
            "%Chng": 0,
            "52Wk-H": 0,
            "52Wk-L": 0,
            "RSI": 0,
            "RSIi": 0,
            "volume": "",
            "22-Pd": "",
            "Consol.": "Range:0%",
            "Breakout": "BO: 0 R: 0",
            "MA-Signal": "",
            "Trend": "",
            "Pattern": "",
            "CCI": 0,
            "FairValue": "-"
        }
        saveDictionary = {
            "Stock": "",
            "LTP": 0,
            "%Chng": 0,
            "52Wk-H": 0,
            "52Wk-L": 0,
            "RSI": 0,
            "RSIi": 0,
            "volume": "",
            "22-Pd": "",
            "Consol.": "Range:0%",
            "Breakout": "BO: 0 R: 0",
            "MA-Signal": "",
            "Trend": "",
            "Pattern": "",
            "CCI": 0,
            "FairValue": "-"
        }
        for prd in periods:
            columns.append(f"LTP{prd}")
            columns.append(f"Growth{prd}")
            screeningDictionary[f"LTP{prd}"] = np.nan
            saveDictionary[f"LTP{prd}"] = np.nan
            screeningDictionary[f"Growth{prd}"] = np.nan
            saveDictionary[f"Growth{prd}"] = np.nan

        screenResults = pd.DataFrame(columns=columns)

        return screeningDictionary, saveDictionary



#!/usr/bin/env python
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import glob
import os
import pickle
import shutil
import tempfile
import pandas as pd
import numpy as np
from halo import Halo
from alive_progress import alive_bar
# from yfinance import shared

from PKDevTools.classes.log import default_logger
from PKDevTools.classes import Archiver
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.MarketHours import MarketHours
from PKDevTools.classes.Committer import Committer
from PKDevTools.classes.SuppressOutput import SuppressOutput
from PKDevTools.classes.PKBackupRestore import start_backup

import pkscreener.classes.Fetcher as Fetcher
from pkscreener.classes.PKTask import PKTask
from pkscreener.classes import Utility, ImageUtility
import pkscreener.classes.ConfigManager as ConfigManager
from pkscreener.classes.PKScheduler import PKScheduler

class PKAssetsManager:
    fetcher = Fetcher.screenerStockDataFetcher()
    configManager = ConfigManager.tools()
    configManager.getConfig(ConfigManager.parser)

    def make_hyperlink(value):
        url = "https://in.tradingview.com/chart?symbol=NSE:{}"
        return '=HYPERLINK("%s", "%s")' % (url.format(ImageUtility.PKImageTools.stockNameFromDecoratedName(value)), value)

    # Save screened results to excel
    def promptSaveResults(sheetName,df_save, defaultAnswer=None,pastDate=None,screenResults=None):
        """
        Tries to save the dataframe output into an excel file.

        It will first try to save to the current-working-directory/results/

        If it fails to save, it will then try to save to Desktop and then eventually into
        a temporary directory.
        """
        data = df_save.copy()
        try:
            data = data.fillna(0)
            data = data.replace([np.inf, -np.inf], 0)
            data = ImageUtility.PKImageTools.removeAllColorStyles(data)
        except KeyboardInterrupt: # pragma: no cover
            raise KeyboardInterrupt
        except Exception as e: # pragma: no cover
            default_logger().debug(e,exc_info=True)
            pass
        try:
            data.reset_index(inplace=True)
            with pd.option_context('mode.chained_assignment', None):
                data["Stock"] = data['Stock'].apply(PKAssetsManager.make_hyperlink)
            data.set_index("Stock", inplace=True)
        except: # pragma: no cover
            pass
        df = data
        isSaved = False
        try:
            if defaultAnswer is None:
                responseLegends = str(
                        OutputControls().takeUserInput(
                            colorText.WARN
                            + f"[>] Do you want to review legends used in the report above? [Y/N](Default:{colorText.END}{colorText.FAIL}N{colorText.END}): ",defaultInput="N"
                        ) or "N"
                    ).upper()
                if "Y" in responseLegends:
                    OutputControls().printOutput(ImageUtility.PKImageTools.getLegendHelpText(table=None).replace("***:",colorText.END+":").replace("***"," " +colorText.FAIL))
                if not PKAssetsManager.configManager.alwaysExportToExcel:
                    response = str(
                        input(
                            colorText.WARN
                            + f"[>] Do you want to save the results in excel file? [Y/N](Default:{colorText.END}{colorText.FAIL}N{colorText.END}): "
                        ) or "N"
                    ).upper()
                else:
                    response = "Y"
            else:
                response = defaultAnswer
        except ValueError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            response = "Y"
        if response is not None and response.upper() != "N":
            pastDateString = f"{pastDate}_to_" if pastDate is not None else ""
            filename = (
                f"PKS_{sheetName.strip()}_"
                + pastDateString
                + PKDateUtilities.currentDateTime().strftime("%d-%m-%y_%H.%M.%S")
                + ".xlsx"
            )
            desktop = os.path.expanduser("~/Desktop")
            # # the above is valid on Windows (after 7) but if you want it in os normalized form:
            desktop = os.path.normpath(os.path.expanduser("~/Desktop"))
            filePath = ""
            try:
                filePath = os.path.join(Archiver.get_user_reports_dir(), filename)
                # Create a Pandas Excel writer using XlsxWriter as the engine.
                writer = pd.ExcelWriter(filePath, engine='xlsxwriter') # openpyxl throws an error exporting % sign.
                # Convert the dataframe to an XlsxWriter Excel object.
                df.to_excel(writer, sheet_name=sheetName[-31:]) # sheetname cannot be beyond 31 character
                # Close the Pandas Excel writer and output the Excel file.
                writer.close()
                df.to_csv(filePath.replace(".xlsx",".csv"))
                isSaved = True
            except KeyboardInterrupt: # pragma: no cover
                raise KeyboardInterrupt
            except Exception as e:  # pragma: no cover
                default_logger().debug(e, exc_info=True)
                OutputControls().printOutput(
                    colorText.FAIL
                    + (
                        "  [+] Error saving file at %s"
                        % filePath
                    )
                    + colorText.END
                )
                try:
                    filePath = os.path.join(desktop, filename)
                    # Create a Pandas Excel writer using XlsxWriter as the engine.
                    writer = pd.ExcelWriter(filePath, engine='xlsxwriter') # openpyxl throws an error exporting % sign.
                    # Convert the dataframe to an XlsxWriter Excel object.
                    df.to_excel(writer, sheet_name=sheetName)
                    # Close the Pandas Excel writer and output the Excel file.
                    writer.close()
                    isSaved = True
                except KeyboardInterrupt: # pragma: no cover
                    raise KeyboardInterrupt
                except Exception as ex:  # pragma: no cover
                    default_logger().debug(ex, exc_info=True)
                    OutputControls().printOutput(
                        colorText.FAIL
                        + (
                            "  [+] Error saving file at %s"
                            % filePath
                        )
                        + colorText.END
                    )
                    try:
                        filePath = os.path.join(tempfile.gettempdir(), filename)
                        # Create a Pandas Excel writer using XlsxWriter as the engine.
                        writer = pd.ExcelWriter(filePath, engine='xlsxwriter') # openpyxl throws an error exporting % sign.
                        # Convert the dataframe to an XlsxWriter Excel object.
                        df.to_excel(writer, sheet_name=sheetName)
                        # Close the Pandas Excel writer and output the Excel file.
                        writer.close()
                        isSaved = True
                    except Exception as ex:  # pragma: no cover
                        pass
            OutputControls().printOutput(
                (colorText.GREEN if isSaved else colorText.FAIL)
                + (("  [+] Results saved to %s" % filePath) if isSaved else "  [+] Failed saving results into Excel file!")
                + colorText.END
            )
            return filePath
        return None

    def afterMarketStockDataExists(intraday=False, forceLoad=False):
        exists, cache_file = Archiver.afterMarketStockDataExists(intraday=intraday,
                                                                 forceLoad=forceLoad,
                                                                 date_suffix=True)
        return exists, cache_file

    @Halo(text='', spinner='dots')
    def saveStockData(stockDict, configManager, loadCount, intraday=False, downloadOnly=False, forceSave=False):
        exists, fileName = PKAssetsManager.afterMarketStockDataExists(
            configManager.isIntradayConfig() or intraday
        )
        outputFolder = Archiver.get_user_data_dir()
        if downloadOnly:
            outputFolder = outputFolder.replace(f"results{os.sep}Data","actions-data-download")
            if not os.path.isdir(outputFolder):
                try:
                    os.makedirs(os.path.dirname(f"{outputFolder}{os.sep}"), exist_ok=True)
                except: # pragma: no cover
                    pass
            configManager.deleteFileWithPattern(rootDir=outputFolder)
        cache_file = os.path.join(outputFolder, fileName)
        if not os.path.exists(cache_file) or forceSave or (loadCount >= 0 and len(stockDict) > (loadCount + 1)):
            try:
                with open(cache_file, "wb") as f:
                    pickle.dump(stockDict.copy(), f, protocol=pickle.HIGHEST_PROTOCOL)
                    OutputControls().printOutput(colorText.GREEN + "=> Done." + colorText.END)
                if downloadOnly:
                    # if "RUNNER" not in os.environ.keys():
                        # copyFilePath = os.path.join(Archiver.get_user_data_dir(), f"copy_{fileName}")
                        # cacheFileSize = os.stat(cache_file).st_size if os.path.exists(cache_file) else 0
                        # if os.path.exists(cache_file) and cacheFileSize >= 1024*1024*40:
                        #     shutil.copy(cache_file,copyFilePath) # copy is the saved source of truth

                    rootDirs = [Archiver.get_user_data_dir(),Archiver.get_user_indices_dir(),outputFolder]
                    patterns = ["*.csv","*.pkl"]
                    for dir in rootDirs:
                        for pattern in patterns:
                            for f in glob.glob(pattern, root_dir=dir, recursive=True):
                                OutputControls().printOutput(colorText.GREEN + f"=> {f}" + colorText.END)
                                if "RUNNER" in os.environ.keys():
                                    Committer.execOSCommand(f"git add {f} -f >/dev/null 2>&1")

            except pickle.PicklingError as e:  # pragma: no cover
                default_logger().debug(e, exc_info=True)
                OutputControls().printOutput(
                    colorText.FAIL
                    + "=> Error while Caching Stock Data."
                    + colorText.END
                )
            except KeyboardInterrupt: # pragma: no cover
                raise KeyboardInterrupt
            except Exception as e:  # pragma: no cover
                default_logger().debug(e, exc_info=True)
        else:
            OutputControls().printOutput(
                colorText.GREEN + "=> Already Cached." + colorText.END
            )
            if downloadOnly:
                OutputControls().printOutput(colorText.GREEN + f"=> {cache_file}" + colorText.END)
        return cache_file

    def had_rate_limit_errors():
        return False
        """Checks if any stored errors are YFRateLimitError."""
        err = "" #",".join(list(shared._ERRORS.values()))
        hitRateLimit = "YFRateLimitError" in err or "Too Many Requests" in err or "429" in err
        if hitRateLimit:
            OutputControls().printOutput(
                colorText.FAIL
                + "  [+] We hit a rate limit error in the previous request(s)!"
                + colorText.END
            )
        return hitRateLimit
    
    @Halo(text='  [+] Downloading fresh data from Data Providers...', spinner='dots')
    def downloadLatestData(stockDict,configManager,stockCodes=[],exchangeSuffix=".NS",downloadOnly=False,numStocksPerIteration=0):
        """
        shared._ERRORS.clear()  # Clear previous errors
        # if numStocksPerIteration == 0:
        # maxParallelProcesses = 17
        numStocksPerIteration = 100 #(int(len(stockCodes)/int(len(stockCodes)/maxParallelProcesses)) if len(stockCodes) >= maxParallelProcesses else len(stockCodes)) + 1
        queueCounter = 0
        iterations = int(len(stockCodes)/numStocksPerIteration) + 1
        tasksList = []
        while queueCounter < iterations:
            stocks = []
            if queueCounter < iterations:
                stocks = stockCodes[numStocksPerIteration* queueCounter : numStocksPerIteration* (queueCounter + 1)]
            else:
                stocks = ["DUMMYStock"]#stockCodes[numStocksPerIteration* queueCounter :]
            fn_args = (stocks, configManager.period, configManager.duration,exchangeSuffix)
            task = PKTask(f"DataDownload-{queueCounter}",long_running_fn=PKAssetsManager.fetcher.fetchStockDataWithArgs,long_running_fn_args=fn_args)
            task.userData = stocks
            if len(stocks) > 0:
                tasksList.append(task)
            queueCounter += 1
        """
        processedStocks = []
        """
        if len(tasksList) > 0:
            # Suppress any multiprocessing errors/warnings
            with SuppressOutput(suppress_stderr=True, suppress_stdout=True):
                PKScheduler.scheduleTasks(tasksList=tasksList, 
                                        label=f"Downloading latest data [{configManager.period},{configManager.duration}] (Total={len(stockCodes)} records in {len(tasksList)} batches){'Be Patient!' if len(stockCodes)> 2000 else ''}",
                                        timeout=(5+2.5*configManager.longTimeout*(4 if downloadOnly else 1)), # 5 sec additional time for multiprocessing setup
                                        minAcceptableCompletionPercentage=(100 if downloadOnly else 100),
                                        showProgressBars=configManager.logsEnabled)
            for task in tasksList:
                if task.result is not None and isinstance(task.result,pd.DataFrame) and not task.result.empty:
                    for stock in task.userData:
                        taskResult = task.result.get(f"{stock}{exchangeSuffix}")
                        if taskResult is not None and isinstance(taskResult,pd.DataFrame) and not taskResult.empty:
                            stockDict[stock] = taskResult.to_dict("split")
                            processedStocks.append(stock)
        """
        leftOutStocks = list(set(stockCodes)-set(processedStocks))
        default_logger().debug(f"Attempted fresh download of {len(stockCodes)} stocks and downloaded {len(processedStocks)} stocks. {len(leftOutStocks)} stocks remaining.")
        return stockDict, leftOutStocks

    def loadStockData(
        stockDict,
        configManager,
        downloadOnly=False,
        defaultAnswer=None,
        retrial=False,
        forceLoad=False,
        stockCodes=[],
        exchangeSuffix=".NS",
        isIntraday = False,
        forceRedownload=False,
        userDownloadOption=None
    ):
        isIntraday = isIntraday or configManager.isIntradayConfig()
        exists, cache_file = PKAssetsManager.afterMarketStockDataExists(
            isIntraday, forceLoad=forceLoad
        )
        initialLoadCount = len(stockDict)
        leftOutStocks = None
        recentDownloadFromOriginAttempted = False
        srcFilePath = os.path.join(Archiver.get_user_data_dir(), cache_file)
        isTrading = PKDateUtilities.isTradingTime() and (PKDateUtilities.wasTradedOn() or not PKDateUtilities.isTodayHoliday()[0])
        if isTrading or not os.path.exists(srcFilePath):
            try:
                from pkbrokers.kite.examples.externals import kite_fetch_save_pickle
                if kite_fetch_save_pickle():
                    default_logger().info("pkl file update succeeded!")
            except Exception as e:
                default_logger().error(f"Error downloading latest file:{e}")
            isTrading = False
        if userDownloadOption is not None and "B" in userDownloadOption: # Backtests
            isTrading = False
        # Check if NSEI data is requested
        if configManager.baseIndex not in stockCodes:
            stockCodes.insert(0,configManager.baseIndex)
        # stockCodes is not None mandates that we start our work based on the downloaded data from yesterday
        if (stockCodes is not None and len(stockCodes) > 0) and (isTrading or downloadOnly):
            recentDownloadFromOriginAttempted = True
            stockDict, leftOutStocks = PKAssetsManager.downloadLatestData(stockDict,configManager,stockCodes,exchangeSuffix=exchangeSuffix,downloadOnly=downloadOnly,numStocksPerIteration=len(stockCodes) if stockCodes is not None else 0)
            if len(leftOutStocks) > int(len(stockCodes)*0.05) and not PKAssetsManager.had_rate_limit_errors():
                # During live market hours, we may not really get additional data if we didn't
                # get it the first time
                # More than 5 % of stocks are still remaining
                stockDict, _ = PKAssetsManager.downloadLatestData(stockDict,configManager,leftOutStocks,exchangeSuffix=exchangeSuffix,downloadOnly=downloadOnly,numStocksPerIteration=len(leftOutStocks) if leftOutStocks is not None else 0)
            # return stockDict
        if downloadOnly or isTrading:
            # We don't want to download from local stale pkl file or stale file at server
            # start_backup()
            return stockDict
        
        default_logger().debug(
            f"Stock data cache file:{cache_file} exists ->{str(exists)}"
        )
        stockDataLoaded = False
        # copyFilePath = os.path.join(Archiver.get_user_data_dir(), f"copy_{cache_file}")
        # if os.path.exists(copyFilePath):
        #     shutil.copy(copyFilePath,srcFilePath) # copy is the saved source of truth
        if os.path.exists(srcFilePath) and not forceRedownload:
            stockDict, stockDataLoaded = PKAssetsManager.loadDataFromLocalPickle(stockDict,configManager, downloadOnly, defaultAnswer, exchangeSuffix, cache_file, isTrading)
        if (
            not stockDataLoaded
            and ("1d" if isIntraday else ConfigManager.default_period)
            == configManager.period
            and ("1m" if isIntraday else ConfigManager.default_duration)
            == configManager.duration
        ) or forceRedownload:
            stockDict, stockDataLoaded = PKAssetsManager.downloadSavedDataFromServer(stockDict,configManager, downloadOnly, defaultAnswer, retrial, forceLoad, stockCodes, exchangeSuffix, isIntraday, forceRedownload, cache_file, isTrading)
        if not stockDataLoaded:
            OutputControls().printOutput(
                colorText.FAIL
                + "  [+] Cache unavailable on pkscreener server, Continuing.."
                + colorText.END
            )
        if not stockDataLoaded and not recentDownloadFromOriginAttempted and not PKAssetsManager.had_rate_limit_errors():
            stockDict, _ = PKAssetsManager.downloadLatestData(stockDict,configManager,stockCodes,exchangeSuffix=exchangeSuffix,downloadOnly=downloadOnly,numStocksPerIteration=len(stockCodes) if stockCodes is not None else 0)
        # See if we need to save stock data
        stockDataLoaded = stockDataLoaded or (len(stockDict) > 0 and (len(stockDict) != initialLoadCount))
        leftOutStocks = list(set(stockCodes)-set(list(stockDict.keys())))
        if len(leftOutStocks) > int(len(stockCodes)*0.05) and not PKAssetsManager.had_rate_limit_errors():
            # More than 5 % of stocks are still remaining
            stockDict, _ = PKAssetsManager.downloadLatestData(stockDict,configManager,leftOutStocks,exchangeSuffix=exchangeSuffix,downloadOnly=downloadOnly,numStocksPerIteration=len(leftOutStocks) if leftOutStocks is not None else 0)
        if stockDataLoaded and downloadOnly:
            PKAssetsManager.saveStockData(stockDict,configManager,initialLoadCount,isIntraday,downloadOnly, forceSave=stockDataLoaded)
        # start_backup()
        return stockDict

    @Halo(text='  [+] Loading data from local cache...', spinner='dots')
    def loadDataFromLocalPickle(stockDict, configManager, downloadOnly, defaultAnswer, exchangeSuffix, cache_file, isTrading):
        stockDataLoaded = False
        srcFilePath = os.path.join(Archiver.get_user_data_dir(), cache_file)

        try:
            with open(srcFilePath, "rb") as f:
                stockData = pickle.load(f)
            if not stockData:
                return stockDict, stockDataLoaded
            if not downloadOnly:
                OutputControls().printOutput(
                    colorText.GREEN
                    + f"\n  [+] Automatically Using [{len(stockData)}] Tickers' Cached Stock Data"
                    + (" due to After-Market hours" if not PKDateUtilities.isTradingTime() else "")
                    + colorText.END
                )
            multiIndex = stockData.keys()
            if isinstance(multiIndex, pd.MultiIndex):
                listStockCodes = sorted(set(multiIndex.get_level_values(0)))
            else:
                listStockCodes = list(stockData.keys())
            if exchangeSuffix and any(exchangeSuffix in code for code in listStockCodes):
                listStockCodes = [x.replace(exchangeSuffix, "") for x in listStockCodes]
            for stock in listStockCodes:
                df_or_dict = stockData.get(stock)
                df_or_dict = df_or_dict.to_dict("split") if isinstance(df_or_dict, pd.DataFrame) else df_or_dict
                existingPreLoadedData = stockDict.get(stock)
                if existingPreLoadedData:
                    if isTrading:
                        for col in ["MF", "FII", "MF_Date", "FII_Date", "FairValue"]:
                            existingPreLoadedData[col] = df_or_dict.get(col)
                        stockDict[stock] = existingPreLoadedData
                    else:
                        stockDict[stock] = {**existingPreLoadedData, **df_or_dict}
                elif not isTrading:
                    stockDict[stock] = df_or_dict
            stockDataLoaded = True
        except (pickle.UnpicklingError, EOFError) as e:
            default_logger().debug(e, exc_info=True)
            OutputControls().printOutput(
                colorText.FAIL + "  [+] Error while Reading Stock Cache." + colorText.END
            )
            if PKAssetsManager.promptFileExists(defaultAnswer=defaultAnswer) == "Y":
                configManager.deleteFileWithPattern()
        except KeyboardInterrupt:
            raise
        return stockDict, stockDataLoaded

    @Halo(text='', spinner='dots')
    def downloadSavedDefaultsFromServer(cache_file):
        fileDownloaded = False
        resp = Utility.tools.tryFetchFromServer(cache_file)
        if resp is not None:
            default_logger().debug(
                    f"Stock data cache file:{cache_file} request status ->{resp.status_code}"
                )
        if resp is not None and resp.status_code == 200:
            contentLength = resp.headers.get("content-length")
            serverBytes = int(contentLength) if contentLength is not None else 0
            KB = 1024
            MB = KB * 1024
            chunksize = MB if serverBytes >= MB else (KB if serverBytes >= KB else 1)
            filesize = int( serverBytes / chunksize)
            if filesize > 40: #Something definitely went wrong. It should be upward of 40bytes
                try:
                    with open(os.path.join(Archiver.get_user_data_dir(), cache_file),"w+",) as f: # .split(os.sep)[-1]
                        f.write(resp.text)
                    fileDownloaded = True
                except: # pragma: no cover
                    pass
        return fileDownloaded

    def downloadSavedDataFromServer(stockDict, configManager, downloadOnly, defaultAnswer, retrial, forceLoad, stockCodes, exchangeSuffix, isIntraday, forceRedownload, cache_file, isTrading):
        stockDataLoaded = False
        resp = Utility.tools.tryFetchFromServer(cache_file)
        if resp is not None:
            default_logger().debug(
                    f"Stock data cache file:{cache_file} request status ->{resp.status_code}"
                )
        if resp is not None and resp.status_code == 200:
            contentLength = resp.headers.get("content-length")
            serverBytes = int(contentLength) if contentLength is not None else 0
            KB = 1024
            MB = KB * 1024
            chunksize = MB if serverBytes >= MB else (KB if serverBytes >= KB else 1)
            filesize = int( serverBytes / chunksize)
            if filesize > 20 and chunksize == MB: # Saved data can't be in KBs. Something definitely went wrong. It should be upward of 40MB
                bar, spinner = Utility.tools.getProgressbarStyle()
                try:
                    f = open(
                            os.path.join(Archiver.get_user_data_dir(), cache_file),
                            "w+b",
                        )  # .split(os.sep)[-1]
                    dl = 0
                    with alive_bar(
                            filesize, bar=bar, spinner=spinner, manual=True
                        ) as progressbar:
                        for data in resp.iter_content(chunk_size=chunksize):
                            dl += 1
                            f.write(data)
                            progressbar(dl / filesize)
                            if dl >= filesize:
                                progressbar(1.0)
                    f.close()
                    with open(
                            os.path.join(Archiver.get_user_data_dir(), cache_file),
                            "rb",
                        ) as f:
                        stockData = pickle.load(f)
                    if len(stockData) > 0:
                        multiIndex = stockData.keys()
                        if isinstance(multiIndex, pd.MultiIndex):
                                # If we requested for multiple stocks from yfinance
                                # we'd have received a multiindex dataframe
                            listStockCodes = multiIndex.get_level_values(0)
                            listStockCodes = sorted(list(filter(None,list(set(listStockCodes)))))
                            if len(listStockCodes) > 0 and len(exchangeSuffix) > 0 and exchangeSuffix in listStockCodes[0]:
                                listStockCodes = [x.replace(exchangeSuffix,"") for x in listStockCodes]
                        else:
                            listStockCodes = list(stockData.keys())
                            if len(listStockCodes) > 0 and len(exchangeSuffix) > 0 and exchangeSuffix in listStockCodes[0]:
                                listStockCodes = [x.replace(exchangeSuffix,"") for x in listStockCodes]
                        for stock in listStockCodes:
                            df_or_dict = stockData.get(stock)
                            df_or_dict = df_or_dict.to_dict("split") if isinstance(df_or_dict,pd.DataFrame) else df_or_dict
                                # This will keep all the latest security data we downloaded
                                # just now and also copy the additional data like, MF/FII,FairValue
                                # etc. data, from yesterday's saved data.
                            try:
                                existingPreLoadedData = stockDict.get(stock)
                                if existingPreLoadedData is not None:
                                    if isTrading:
                                            # Only copy the MF/FII/FairValue data and leave the stock prices as is.
                                        cols = ["MF", "FII","MF_Date","FII_Date","FairValue"]
                                        for col in cols:
                                            existingPreLoadedData[col] = df_or_dict.get(col)
                                        stockDict[stock] = existingPreLoadedData
                                    else:
                                        stockDict[stock] = df_or_dict | existingPreLoadedData
                                else:
                                    if not isTrading:
                                        stockDict[stock] = df_or_dict
                            except: # pragma: no cover
                                    # Probably, the "stock" got removed from the latest download
                                    # and so, was not found in stockDict
                                continue
                        stockDataLoaded = True
                        # copyFilePath = os.path.join(Archiver.get_user_data_dir(), f"copy_{cache_file}")
                        # srcFilePath = os.path.join(Archiver.get_user_data_dir(), cache_file)
                        # if os.path.exists(copyFilePath) and os.path.exists(srcFilePath):
                        #     shutil.copy(copyFilePath,srcFilePath) # copy is the saved source of truth
                        # if not os.path.exists(copyFilePath) and os.path.exists(srcFilePath): # Let's make a copy of the original one
                        #     shutil.copy(srcFilePath,copyFilePath)
                        # Remove the progress bar now!
                        OutputControls().moveCursorUpLines(1)
                except KeyboardInterrupt: # pragma: no cover
                    raise KeyboardInterrupt
                except Exception as e:  # pragma: no cover
                    default_logger().debug(e, exc_info=True)
                    f.close()
                    OutputControls().printOutput("[!] Download Error - " + str(e))
            else:
                default_logger().debug(
                        f"Stock data cache file:{cache_file} on server has length ->{filesize} {'Mb' if chunksize >= MB else ('Kb' if chunksize >= KB else 'bytes')}"
                    )
            if not retrial and not stockDataLoaded:
                # Don't try for more than once.
                stockDict = PKAssetsManager.loadStockData(
                        stockDict,
                        configManager,
                        downloadOnly,
                        defaultAnswer,
                        retrial=True,
                        forceLoad=forceLoad,
                        stockCodes=stockCodes,
                        exchangeSuffix=exchangeSuffix,
                        isIntraday = isIntraday,
                        forceRedownload=forceRedownload
                    )
                
        return stockDict,stockDataLoaded

    # Save screened results to excel
    def promptFileExists(cache_file="stock_data_*.pkl", defaultAnswer=None):
        try:
            if defaultAnswer is None:
                response = str(
                    input(
                        colorText.WARN
                        + "[>] "
                        + cache_file
                        + " already exists. Do you want to replace this? [Y/N] (Default: Y): "
                ) or "Y").upper()
            else:
                response = defaultAnswer
        except ValueError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            pass
        return "Y" if response != "N" else "N"



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import os
import datetime
import math
import pandas as pd
import textwrap
import random
import platform
import warnings

warnings.simplefilter("ignore", DeprecationWarning)
warnings.simplefilter("ignore", FutureWarning)

from PIL import Image, ImageDraw, ImageFont
import PIL.Image
PIL.Image.MAX_IMAGE_PIXELS = None
from halo import Halo

from PKDevTools.classes import Archiver
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.PKDateUtilities import PKDateUtilities

from pkscreener.classes import Utility, ConfigManager
from pkscreener.classes.Utility import artText, marketStatus
import pkscreener.classes.Fetcher as Fetcher

# Class for managing misc image utility methods
class PKImageTools:
    fetcher = Fetcher.screenerStockDataFetcher()
    configManager = ConfigManager.tools()
    configManager.getConfig(ConfigManager.parser)

    def getsize_multiline(font,srcText,x=0,y=0):
        zeroSizeImage = Image.new('RGB',(0, 0), (0,0,0))
        zeroDraw = ImageDraw.Draw(zeroSizeImage)
        # zeroDraw = ImageDraw.Draw(zeroSizeImage)
        bbox = zeroDraw.multiline_textbbox((x,y),srcText,font)
        # Calculate width and height from the bounding box
        width = bbox[2] - bbox[0]
        height = bbox[3] - bbox[1]
        # default_logger().debug(f"Text size for text: {srcText}: {width}x{height}")
        return width, height

    def getsize(font,srcText,x=0,y=0):
        left, top, bottom, right = font.getbbox(srcText)
        return right - left, bottom - top
    
    def addQuickWatermark(sourceImage:Image, xVertical=None, dataSrc="", dataSrcFontSize=10):
        width, height = sourceImage.size
        watermarkText = f"© {datetime.date.today().year} pkjmesra | PKScreener"
        message_length = len(watermarkText)
        # load font (tweak ratio based on a particular font)
        FONT_RATIO = 1.5
        DIAGONAL_PERCENTAGE = .85
        DATASRC_FONTSIZE = dataSrcFontSize
        dataSrc = f"Src: {dataSrc}"
        diagonal_length = int(math.sqrt((width**2) + (height**2)))
        diagonal_to_use = diagonal_length * DIAGONAL_PERCENTAGE
        height_to_use = height * DIAGONAL_PERCENTAGE
        font_size = int(diagonal_to_use / (message_length / FONT_RATIO))
        font_size_vertical = int(height_to_use / (message_length / FONT_RATIO))
        fontPath = PKImageTools.setupReportFont()
        font = ImageFont.truetype(fontPath, font_size)
        font_vertical = ImageFont.truetype(fontPath, font_size_vertical)
        #font = ImageFont.load_default() # fallback

        # watermark 1
        opacity = int(256 * .6)
        _,_,mark_width, mark_height = font.getbbox(watermarkText)
        watermark = Image.new('RGBA', (mark_width, mark_height), (0, 0, 0, 0))
        draw = ImageDraw.Draw(watermark)
        draw.text((0, 0), text=watermarkText, font=font, fill=(128, 128, 128, opacity))
        angle = math.degrees(math.atan(height/width))
        watermark_diag = watermark.rotate(angle, expand=1)
        
        _,_,mark_width_ver, mark_height_ver = font_vertical.getbbox(watermarkText)
        watermark_ver = Image.new('RGBA', (mark_width_ver, mark_height_ver), (0, 0, 0, 0))
        draw = ImageDraw.Draw(watermark_ver)
        draw.text((0, 0), text=watermarkText, font=font_vertical, fill=(128, 128, 128, opacity))
        watermark_vertical = watermark_ver.rotate(90, expand=1)

        # merge
        try:
            logo_wm_path = os.path.join(Archiver.get_user_outputs_dir().replace("results","pkscreener"),"LogoWM.txt")
            if not os.path.isfile(logo_wm_path):
                resp = Utility.tools.tryFetchFromServer(cache_file="LogoWM.png",directory="screenshots/logos",hideOutput=True)
                with open(logo_wm_path,"wb",) as f:
                    f.write(resp.content)
            logo_img = Image.open(logo_wm_path,formats=["PNG"]).convert('LA')
            # logo_img = logo_img.resize((min(width,height),min(width,height)), Image.LANCZOS, reducing_gap=2)
            lx, ly = logo_img.size
            plx = int((width - lx)/4)
            ply = int((height - ly)/3)
            sourceImage.paste(logo_img, (plx, ply), logo_img)
        except KeyboardInterrupt: # pragma: no cover
            raise KeyboardInterrupt
        except Exception as e: # pragma: no cover
            default_logger().debug(e,exc_info=True)
            pass

        wx, wy = watermark_diag.size
        px = int((width - wx)/2)
        py = int((height - wy)/2)
        wxv, wyv = watermark_vertical.size
        pxv =  int((width - wxv)/12) if xVertical is None else xVertical
        pyv= int((height - wyv)/2)
        sourceImage.paste(watermark_diag, (px, py, px + wx, py + wy), watermark_diag)
        sourceImage.paste(watermark_vertical, (pxv, pyv, pxv + wxv, pyv + wyv), watermark_vertical)
        
        # Draw the data sources
        dataSrcFont = ImageFont.truetype(fontPath, DATASRC_FONTSIZE)
        dataSrc_width, dataSrc_height = PKImageTools.getsize_multiline(font=dataSrcFont,srcText=dataSrc)
        draw = ImageDraw.Draw(sourceImage)
        draw.text((width-dataSrc_width, height-dataSrc_height-2), text=dataSrc, font=dataSrcFont, fill=(128, 128, 128, opacity))
        # sourceImage.show()
        return sourceImage
    
    def removeAllColorStyles(styledText):
        styles = [
            colorText.HEAD,
            colorText.END,
            colorText.BOLD,
            colorText.UNDR,
            colorText.BLUE,
            colorText.GREEN,
            colorText.BRIGHTGREEN,
            colorText.WARN,
            colorText.BRIGHTYELLOW,
            colorText.FAIL,
            colorText.BRIGHTRED,
            colorText.WHITE,
        ]
        if isinstance(styledText,pd.DataFrame):
            styledTextCopy = styledText.copy()
            with pd.option_context('mode.chained_assignment', None):
                for col in styledTextCopy.columns:
                    for style in styles:
                        try:
                            styledTextCopy[col] = styledTextCopy[col].astype(str).str.replace(style,"")
                        except: # pragma: no cover
                            pass
            return styledTextCopy
        elif isinstance(styledText,str):
            cleanedUpStyledValue = str(styledText)
            for style in styles:
                cleanedUpStyledValue = cleanedUpStyledValue.replace(style, "")
            return cleanedUpStyledValue
        else:
            return styledText

    def getCellColors(cellStyledValue="", defaultCellFillColor="black"):
        otherStyles = [colorText.HEAD, colorText.BOLD, colorText.UNDR]
        mainStyles = [
            colorText.BLUE,
            colorText.GREEN,
            colorText.BRIGHTGREEN,
            colorText.WARN,
            colorText.BRIGHTYELLOW,
            colorText.FAIL,
            colorText.BRIGHTRED,
            colorText.WHITE,
        ]
        colorsDict = {
            colorText.BLUE: "blue",
            colorText.BRIGHTGREEN: "darkgreen",
            colorText.GREEN: "green"
            if defaultCellFillColor == "black"
            else "lightgreen",
            colorText.WARN: "darkorange"
            if defaultCellFillColor == "black"
            else "yellow",
            colorText.BRIGHTYELLOW: "darkyellow",
            colorText.FAIL: "red",
            colorText.BRIGHTRED : "darkred",
            colorText.WHITE: "white" 
            if defaultCellFillColor == "white"
            else "black",
        }
        cleanedUpStyledValues = []
        cellFillColors = []
        cleanedUpStyledValue = str(cellStyledValue)
        prefix = ""
        for style in otherStyles:
            cleanedUpStyledValue = cleanedUpStyledValue.replace(style, "")
        # Find how many different colors are used for the cell value
        coloredStyledValues = cleanedUpStyledValue.split(colorText.END)
        for cleanedUpStyledValue in coloredStyledValues:
            cleanedUpStyledValue = cleanedUpStyledValue.replace(colorText.END,"")
            if cleanedUpStyledValue.strip() in ["", ",","/"]:
                if len(cleanedUpStyledValues) > 0:
                    cleanedUpStyledValues[len(cleanedUpStyledValues)-1] = f"{cleanedUpStyledValues[len(cleanedUpStyledValues)-1]}{cleanedUpStyledValue}"
                else:
                    prefix = cleanedUpStyledValue
            else:
                for style in mainStyles:
                    if style in cleanedUpStyledValue:
                        cellFillColors.append(colorsDict[style])
                        for style in mainStyles:
                            cleanedUpStyledValue = cleanedUpStyledValue.replace(style, "")
                        cleanedUpStyledValues.append(prefix + cleanedUpStyledValue)
                        prefix = ""
                
        if len(cellFillColors) == 0:
            cellFillColors = [defaultCellFillColor]
        if len(cleanedUpStyledValues) == 0:
            cleanedUpStyledValues = [str(cellStyledValue)]
        return cellFillColors, cleanedUpStyledValues

    @Halo(text='', spinner='dots')
    def tableToImage(
        table,
        styledTable,
        filename,
        label,
        backtestSummary=None,
        backtestDetail=None,
        addendum=None,
        addendumLabel=None,
        summaryLabel = None,
        detailLabel = None,
        legendPrefixText = ""
    ):
        if "PKDevTools_Default_Log_Level" not in os.environ.keys():
            if (("RUNNER" in os.environ.keys() and os.environ["RUNNER"] == "LOCAL_RUN_SCANNER")):
                return
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        ART_FONT_SIZE = 30
        STD_FONT_SIZE = 60
        try:
            # First 4 lines are headers. Last 1 line is bottom grid line
            fontPath = PKImageTools.setupReportFont()
            artfont = ImageFont.truetype(fontPath, ART_FONT_SIZE)
            stdfont = ImageFont.truetype(fontPath, STD_FONT_SIZE)
            
            bgColor, gridColor, artColor, menuColor = PKImageTools.getDefaultColors()

            dfs_to_print = [styledTable, backtestSummary, backtestDetail]
            unstyled_dfs = [table, backtestSummary, backtestDetail]
            reportTitle = f"  [+] As of {PKDateUtilities.currentDateTime().strftime('%d-%m-%y %H.%M.%S')} IST > You chose {label}"
            titleLabels = [
                f"  [+] Scan results for {label} :",
                summaryLabel if summaryLabel is not None else "  [+] For chosen scan, summary of correctness from past: [Example, 70% of (100) under 1-Pd, means out of 100 stocks that were in the scan result in the past, 70% of them gained next day.)",
                detailLabel if detailLabel is not None else "  [+] 1 to 30 period gain/loss % for matching stocks on respective date from earlier predictions:[Example, 5% under 1-Pd, means the stock price actually gained 5% the next day from given date.]",
            ]

            artfont_arttext_width, artfont_arttext_height = PKImageTools.getsize_multiline(font=artfont,srcText=artText+ f"{marketStatus()}")
            stdFont_oneLinelabel_width, stdFont_oneLinelabel_height = PKImageTools.getsize_multiline(font=stdfont,srcText=label)
            stdFont_scanResulttext_width, stdFont_scanResulttext_height = PKImageTools.getsize_multiline(font=stdfont,srcText=table) if len(table) > 0 else (0,0)
            unstyled_backtestsummary = PKImageTools.removeAllColorStyles(backtestSummary)
            unstyled_backtestDetail = PKImageTools.removeAllColorStyles(backtestDetail)
            stdFont_backtestSummary_text_width,stdFont_backtestSummary_text_height= PKImageTools.getsize_multiline(font=stdfont,srcText=unstyled_backtestsummary) if (unstyled_backtestsummary is not None and len(unstyled_backtestsummary) > 0) else (0,0)
            stdFont_backtestDetail_text_width, stdFont_backtestDetail_text_height = PKImageTools.getsize_multiline(font=stdfont, srcText=unstyled_backtestDetail) if (unstyled_backtestDetail is not None and len(unstyled_backtestDetail) > 0) else (0,0)
            artfont_scanResultText_width, _ = PKImageTools.getsize_multiline(font=artfont,srcText=table) if len(table) > 0 else (0,0)
            artfont_backtestSummary_text_width, _ = PKImageTools.getsize_multiline(font= artfont,srcText=backtestSummary) if (backtestSummary is not None and len(backtestSummary)) > 0 else (0,0)
            stdfont_addendumtext_height = 0
            stdfont_addendumtext_width = 0
            if addendum is not None and len(addendum) > 0:
                unstyled_addendum = PKImageTools.removeAllColorStyles(addendum)
                stdfont_addendumtext_width , stdfont_addendumtext_height = PKImageTools.getsize_multiline(font=stdfont,srcText=unstyled_addendum)
                titleLabels.append(addendumLabel)
                dfs_to_print.append(addendum)
                unstyled_dfs.append(unstyled_addendum)

            repoText = PKImageTools.getRepoHelpText(table,backtestSummary)
            artfont_repotext_width, artfont_repotext_height = PKImageTools.getsize_multiline(font=artfont,srcText=repoText)
            legendText = legendPrefixText + PKImageTools.getLegendHelpText(table,backtestSummary)
            _, artfont_legendtext_height = PKImageTools.getsize_multiline(font=artfont,srcText=legendText)
            column_separator = "|"
            line_separator = "+"
            stdfont_sep_width, _ = PKImageTools.getsize_multiline(font=stdfont, srcText=column_separator)

            startColValue = 100
            xVertical = startColValue
            rowPixelRunValue = 9
            im_width = max(
                artfont_arttext_width,
                stdFont_oneLinelabel_width,
                stdFont_scanResulttext_width,
                stdFont_backtestSummary_text_width,
                stdFont_backtestDetail_text_width,
                artfont_repotext_width,
                artfont_scanResultText_width,
                artfont_backtestSummary_text_width,
                stdfont_addendumtext_width
            ) + int(startColValue * 2)
            im_height = int(
                        artfont_arttext_height # Always
                        + 3*stdFont_oneLinelabel_height # Title label # Always
                        + stdFont_scanResulttext_height + (stdFont_oneLinelabel_height if int(stdFont_scanResulttext_height) > 0 else 0)
                        + stdFont_backtestSummary_text_height + (stdFont_oneLinelabel_height if int(stdFont_backtestSummary_text_height) > 0 else 0)
                        + stdFont_backtestDetail_text_height + (stdFont_oneLinelabel_height if int(stdFont_backtestDetail_text_height) > 0 else 0)
                        + artfont_repotext_height # Always
                        + artfont_legendtext_height # Always
                        + stdfont_addendumtext_height + (stdFont_oneLinelabel_height if int(stdfont_addendumtext_height) > 0 else 0)
                    )
            im = Image.new("RGB",(im_width,im_height),bgColor)
            draw = ImageDraw.Draw(im)
            # artwork
            draw.text((startColValue, rowPixelRunValue), artText+ f"{PKImageTools.removeAllColorStyles(marketStatus())}", font=artfont, fill=artColor)
            rowPixelRunValue += artfont_arttext_height + 1
            # Report title
            # reportTitle = tools.wrapFitLegendText(table,backtestSummary, reportTitle)
            draw.text((startColValue, rowPixelRunValue), reportTitle, font=stdfont, fill=menuColor)
            rowPixelRunValue += stdFont_oneLinelabel_height + 1
            counter = -1
            for df in dfs_to_print:
                counter += 1
                colPixelRunValue = startColValue
                try:
                    if df is None or len(df) == 0:
                        continue
                except: # pragma: no cover
                    continue
                # selected menu options and As of DateTime
                draw.text(
                    (colPixelRunValue, rowPixelRunValue),
                    titleLabels[counter],
                    font=stdfont,
                    fill=menuColor,
                )
                rowPixelRunValue += stdFont_oneLinelabel_height
                unstyledLines = unstyled_dfs[counter].splitlines()
                lineNumber = 0
                screenLines = df.splitlines()
                for line in screenLines:
                    _, stdfont_line_height = PKImageTools.getsize_multiline(font=stdfont, srcText=line)
                    # Print the row separators
                    if not (line.startswith(column_separator)):
                        draw.text(
                            (colPixelRunValue, rowPixelRunValue),
                            line,
                            font=stdfont,
                            fill=gridColor,
                        )
                        rowPixelRunValue += stdfont_line_height + 1
                    else: # if (line.startswith(column_separator)):
                        # Print the row contents
                        columnNumber = 0
                        valueScreenCols = line.split(column_separator)
                        try:
                            del valueScreenCols[0] # Remove the empty column header at the first position
                            del valueScreenCols[-1] # Remove the empty column header at the last position
                        except KeyboardInterrupt: # pragma: no cover
                            raise KeyboardInterrupt
                        except Exception as e:# pragma: no cover
                            default_logger().debug(e, exc_info=True)
                            draw.text(
                                (colPixelRunValue, rowPixelRunValue),
                                line,
                                font=stdfont,
                                fill=gridColor,
                            )
                            lineNumber = lineNumber - 1
                            pass
                        # Print each colored value of each cell as we go over each row
                        for val in valueScreenCols:
                            if lineNumber >= len(unstyledLines):
                                continue
                            # Draw the column separator first
                            draw.text(
                                (colPixelRunValue, rowPixelRunValue),
                                column_separator,
                                font=stdfont,
                                fill=gridColor,
                            )
                            colPixelRunValue = colPixelRunValue + stdfont_sep_width
                            unstyledLine = unstyledLines[lineNumber]
                            cellStyles, cellCleanValues = PKImageTools.getCellColors(
                                val, defaultCellFillColor=gridColor
                            )
                            valCounter = 0
                            for style in cellStyles:
                                cleanValue = cellCleanValues[valCounter]
                                valCounter += 1
                                if columnNumber == 0 and len(cleanValue.strip()) > 0:
                                    if column_separator in unstyledLine:
                                        cleanValue = unstyledLine.split(column_separator)[1]
                                    if "\\" in cleanValue:
                                        cleanValue = cleanValue.split("\\")[-1]
                                    # style = style if "%" in cleanValue else gridColor
                                if bgColor == "white" and style == "yellow":
                                    # Yellow on a white background is difficult to read
                                    style = "blue"
                                elif bgColor == "black" and style == "blue":
                                    # blue on a black background is difficult to read
                                    style = "yellow"
                                col_width, _ = PKImageTools.getsize_multiline(font=stdfont, srcText=cleanValue)
                                draw.text(
                                    (colPixelRunValue, rowPixelRunValue),
                                    cleanValue,
                                    font=stdfont,
                                    fill=style,
                                )
                                colPixelRunValue = colPixelRunValue + col_width
                                if columnNumber == 0:
                                    xVertical = int(columnNumber/2)

                            columnNumber = columnNumber + 1
                        if len(valueScreenCols) > 0:
                            # Close the row with the separator
                            draw.text(
                                (colPixelRunValue, rowPixelRunValue),
                                column_separator,
                                font=stdfont,
                                fill=gridColor,
                            )
                            colPixelRunValue = startColValue
                        rowPixelRunValue +=  stdfont_line_height + 1
                    lineNumber = lineNumber + 1
                rowPixelRunValue += stdFont_oneLinelabel_height
            
            # Repo text
            draw.text(
                (colPixelRunValue, rowPixelRunValue + 1),
                repoText,
                font=artfont,
                fill=menuColor,
            )
            # Legend text
            rowPixelRunValue += 2 * stdFont_oneLinelabel_height + 20
            legendLines = legendText.splitlines()
            legendSeperator = "***"
            col_width_sep, _ = PKImageTools.getsize_multiline(font=artfont, srcText=legendSeperator)
            for line in legendLines:
                colPixelRunValue = startColValue
                _, artfont_line_height = PKImageTools.getsize_multiline(font=artfont, srcText=line)
                lineitems = line.split(legendSeperator)
                red = True
                for lineitem in lineitems:
                    if lineitem == "" or not red:
                        draw.text(
                            (colPixelRunValue, rowPixelRunValue),
                            legendSeperator,
                            font=artfont,
                            fill=gridColor,
                        )
                        colPixelRunValue += col_width_sep + 1
                    style = "red" if not red else gridColor
                    red = not red
                    lineitem = lineitem.replace(": ","***: ")
                    draw.text(
                        (colPixelRunValue, rowPixelRunValue),
                        lineitem,
                        font=artfont,
                        fill=style,
                    )
                    col_width, _ = PKImageTools.getsize_multiline(font=artfont, srcText=lineitem)
                    # Move to the next text in the same line
                    colPixelRunValue += col_width + 1
                    
                # Let's go to the next line
                rowPixelRunValue += artfont_line_height + 1

            im = im.resize((int(im.size[0]*PKImageTools.configManager.telegramImageCompressionRatio),int(im.size[1]*PKImageTools.configManager.telegramImageCompressionRatio)), Image.LANCZOS, reducing_gap=2)
            im = PKImageTools.addQuickWatermark(im,xVertical,dataSrc="Yahoo!finance; Morningstar, Inc; National Stock Exchange of India Ltd;TradingHours.com;",dataSrcFontSize=ART_FONT_SIZE)
            im.save(filename, format=PKImageTools.configManager.telegramImageFormat, bitmap_format=PKImageTools.configManager.telegramImageFormat, optimize=True, quality=int(PKImageTools.configManager.telegramImageQualityPercentage))
        except KeyboardInterrupt: # pragma: no cover
            raise KeyboardInterrupt
        except Exception as e: # pragma: no cover
            default_logger().debug(e, exc_info=True)
        # if 'RUNNER' not in os.environ.keys() and 'PKDevTools_Default_Log_Level' in os.environ.keys():
        # im.show()

    def wrapFitLegendText(table=None, backtestSummary=None, legendText=None):
        if legendText is None or len(legendText) < 1:
            return legendText
        wrapper = textwrap.TextWrapper(
            width=2
            * int(
                len(table.split("\n")[0])
                if (table is not None and len(table) > 0)
                else (len(backtestSummary.split("\n")[0]) if backtestSummary is not None else 80)
            )
        )
        if wrapper.width <= 0:
            return legendText
        word_list = wrapper.wrap(text=legendText)
        legendText_new = ""
        for ii in word_list[:-1]:
            legendText_new = legendText_new + ii + "\n"
        legendText_new += word_list[-1]
        legendText = legendText_new
        return legendText

    def getDefaultColors():
        artColors = ["blue", "indigo", "green", "red", "yellow","orange","violet"]
        bgColor = "white" if PKDateUtilities.currentDateTime().day % 2 == 0 else "black"
        gridColor = "black" if bgColor == "white" else "white"
        artColor = random.choice(artColors[3:]) if bgColor == "black" else random.choice(artColors[:3])
        menuColor = "red"
        return bgColor,gridColor,artColor,menuColor

    def setupReportFont():
        fontURL = "https://raw.githubusercontent.com/pkjmesra/pkscreener/main/pkscreener/courbd.ttf"
        fontFile = fontURL.split("/")[-1]
        bData, fontPath, _ = Archiver.findFile(fontFile)
        if bData is None:
            resp = PKImageTools.fetcher.fetchURL(fontURL, stream=True)
            if resp is not None:
                with open(fontPath, "wb") as f:
                    for chunk in resp.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
            else:
                path1 = os.path.join(Archiver.get_user_outputs_dir().replace("results","pkscreener"),"courbd.ttf")
                path2 = os.path.join(os.getcwd(),"courbd.ttf")
                if os.path.isfile(path1):
                    fontPath = path1
                elif os.path.isfile(path2):
                    fontPath = path2
                else:
                    if "Windows" in platform.system():
                        fontPath = "arial.ttf"
                    elif "Darwin" in platform.system():
                        fontPath = "/System/Library/Fonts/Keyboard.ttf"
                    else:
                        fontPath = "/usr/share/fonts/truetype/freefont/FreeMono.ttf"
        return fontPath

    def getLegendHelpText(table=None,backtestSummary=None):
        legendText = "\n***1.Stock***: This is the NSE symbol/ticker for a company. Stocks that are NOT stage two, are coloured red.***2.Consol.***: It shows the price range in which stock is trading for the last 22 trading sessions(22 trading sessions per month)***3.Breakout(22Prds)***: The BO is Breakout level based on last 22 sessions. R is the resistance level (if available)."
        legendText = f"{legendText} An investor should consider both BO & R level to analyse entry / exits in their trading lessons. If the BO value is green, it means the stock has already broken out (is above BO level). If BO is in red, it means the stock is yet to break out.***4.LTP***: This is the last/latest trading/closing price of the given stock on a given date at NSE. The LTP in green/red means the"
        legendText = f"{legendText} stock price has increased / decreased since last trading session. (1.5%, 1.3%,1.8%) with LTP shows the stock price rose by 1.5%, 1.3% and 1.8% in the last 1, 2 and 3 trading sessions respectively.***5.%Chng***: This is the change(rise/fall in percentage) in closing/trading price from the previous trading session's closing price. Green means that price rose from the previous"
        legendText = f"{legendText} trading session. Red means it fell.***6.volume***: This shows the relative volume in the most recent trading day /today with respect to last 20 trading periods moving average of Volume. For example, 8.5x would mean today's volume so far is 8.5 times the average volume traded in the last 20 trading sessions. Volume in green means that volume for the date so far has been at"
        legendText = f"{legendText} least 2.5 times more than the average volume of last 20 sessions. If the volume is in red, it means the given date's volume is less than 2.5 times the avg volume of the last 20 sessions.***7.MA-Signal***: It shows the price trend of the given stock by analyzing various 50-200 moving/exponential averages crossover strategies. Perform a Google search for the shown MA-Signals"
        legendText = f"{legendText} to learn about them more. If it is in green, the signal is bullish. Red means bearish.***8.RSI-or-RSI/i***: Relative Strength Index is a momentum index which describes 14-period relative strength at the given price. Generally, below 30 is considered oversold and above 80 is considered overbought. When RSI/i has value, say, 80/41, it means that the daily RSI value is 80 while"
        legendText = f"{legendText} the 1-minute intraday RSI is 41.***9.Trend(22Prds)***:  This describes the average trendline computed based on the last 22 trading sessions. Their strength is displayed depending on the steepness of the trendlines. (Strong / Weak) Up / Down shows how high/low the demand is respectively. A Sideways trend is the horizontal price movement that occurs when the forces of supply"
        legendText = f"{legendText} and demand are nearly equal. T:▲ or T:▼ shows the general moving average uptrend/downtrend from a 200 day MA perspective"
        legendText = f"{legendText} if the current 200DMA is more/less than the last 20/80/100 days' 200DMA. Similarly, t:▲ or t:▼ shows for 50DMA based on 9/14/20 days' 50DMA trend. MFI:▲ or MFI:▼ shows"
        legendText = f"{legendText} if the overall top 5 mutual funds and top 5 institutional investors ownership went up or down on the closing of the last month.***10.Pattern***:This shows if the chart or the candle (from the candlestick chart) is"
        legendText = f"{legendText} forming any known pattern in the recent timeframe or as per the selected screening options. Do a google search for the shown pattern names to learn.***11.CCI***: The Commodity Channel Index (CCI) is a technical indicator that measures the difference between the current price and the historical average price of the given stock. Generally below '- 100' is considered oversold"
        legendText = f"{legendText} and above 100 is considered overbought. If the CCI is < '-100' or CCI is > 100 and the Trend(22Prds) is Strong/Weak Up, it is shown in green. Otherwise it's in red.***12.1-Pd/2-Pd-etc.***: 60.29% of (413) under 1-Pd in green shows that the given scan option was correct 60.23% of the times for 413 stocks that scanner found in the last 22 trading sessions under the same scan"
        legendText = f"{legendText} options. Similarly, 61.69 % of (154) in green under 22-Pd, means we found that 61.56% of 154 stocks (~95 stocks) prices found under the same scan options increased in 22 trading periods. 57.87% of (2661) under 'Overall' means that over the last 22 trading sessions we found 2661 stock instances under the same scanning options (for example, Momentum Gainers), out of which 57.87%"
        legendText = f"{legendText} of the stock prices increased in one or more of the last 1 or 2 or 3 or 4 or 5 or 10 or 22 or 22 trading sessions. If you want to see by what percent the prices increased, you should see the details.***13.1-to-30-period-gain/loss%***: 4.17% under 1-Pd in green in the gain/loss table/grid means the stock price increased by 4.17% in the next 1 trading session. If this is in"
        legendText = f"{legendText} red, example, -5.67%, it means the price actually decreased by 5.67%. Gains are in green and losses are in red in this grid. The Date column has the date(s) on which that specific stock was found under the chosen scan options in the past 22 trading sessions.***14.52Wk-H/L***: These have 52 weeks high/low prices and will be shown in red, green or yellow based on how close the"
        legendText = f"{legendText} price is to the 52 wk high/low value.If the 52 week high/low value is within 10% of LTP:Yellow, LTP is above 52 week high:Green. If the LTP is below 90% of 52 week high:Red.***15.1-Pd-%***: Shows the 1 period gain in percent from the given date. Similarly 2-Pd-%, 3-Pd-% etc shows 2 day, 3 days gain etc.***16.1-Pd-10k***: Shows 1 period/day portfolio value if you would"
        legendText = f"{legendText} have invested 10,000 on the given date.***17.[T][_trend_]***: [T] is for Trends followed by the trend name in the filter.***18.[BO]***: This Shows the Breakout filter value from the backtest reports and will be available only if 'showpaststrategydata' configuration is turned on.***19.[P]***: [P] shows pattern name.***20.MFI***: Top 5 Mutual fund ownership and "
        legendText = f"{legendText} top 5 Institutional investor ownership status as on the last day of the last month, based on analysis from Morningstar.***21.FairValue***: Morningstar Fair value of a given stock as of last trading day as determined by 3rd party analysis based on fundamentals.***22.MCapWt%***: This shows the market-cap weighted portfolio weight to consider investing. "
        legendText = f"{legendText} ***23.Block/Bulk/Short Deals***: Ⓑ : Bulk Deals,Ⓛ: Block Deals,Ⓢ: Short deals. (B) indicates Buy, (S) indicates Sell. (1M) or (1K) indicates the quantity in million/kilo(thousand).\n"
        legendText = PKImageTools.wrapFitLegendText(table=table,backtestSummary=backtestSummary, legendText=legendText)
        # legendText = legendText.replace("***:", colorText.END + colorText.WHITE)
        # legendText = legendText.replace("***", colorText.END + colorText.FAIL)
        # return colorText.WHITE + legendText + colorText.END
        return legendText

    def getRepoHelpText(table,backtestSummary):
        repoText = f"Source: https://GitHub.com/pkjmesra/pkscreener/  | © {datetime.date.today().year} pkjmesra | Telegram: https://t.me/PKScreener |"
        disclaimer = f"The author is NOT a financial advisor and is NOT SEBI registered. This report is for learning/analysis purposes ONLY. Author assumes no responsibility or liability for any errors or omissions in this report or repository, or gain/loss bearing out of this analysis. The user MUST take advise ONLY from registered SEBI financial advisors only."
        repoText = f"{repoText}\n{PKImageTools.wrapFitLegendText(table,backtestSummary,disclaimer)}"
        repoText = f"{repoText}\n  [+] Understanding this report:\n\n"
        return repoText

    def roundOff(value,places):
        roundValue = value
        try:
            newValue = PKImageTools.removeAllColorStyles(str(roundValue))
            newValue = newValue.replace("%","").replace("x","")
            roundValue = round(float(newValue),places)
            if places == 0:
                roundValue = int(roundValue)
            roundValue = str(value).replace(str(newValue),str(roundValue))
        except: # pragma: no cover
            pass
        return roundValue
    
    def stockNameFromDecoratedName(stockName):
        if stockName is None:
            raise TypeError
        cleanName = PKImageTools.removeAllColorStyles(stockName.replace("\x1B]8;;",""))
        decoratedParts = cleanName.split("\x1B\\")
        if len(decoratedParts) > 1:
            cleanName = decoratedParts[1]
        return cleanName
